
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_multitask.multithreading.functionthreaded import FunctionThreaded
from holado_core.common.handlers.redo import Redo
from holado_core.common.exceptions.timeout_exception import TimeoutException
from holado_core.common.handlers.wait import WaitIsTrueOnEvent
from holado_rabbitmq.tools.rabbitmq.rabbitmq_client import RMQClient,\
    RMQPublisher
import queue
import threading
from holado_core.common.tools.tools import Tools


logger = logging.getLogger(__name__)


if RMQClient.is_available():
    import pika
    from pika.exceptions import ConnectionClosedByClient
    from pika.exchange_type import ExchangeType
    
    class RMQChannel(pika.channel.Channel):
        """
        Add features to pika.channel.Channel:
          - Confirm delivery
          - Automatically wait exchange, queue and bind are OK, like with a blocking connection
        """
        def __init__(self, connection, channel_number, on_open_callback):
            super().__init__(connection, channel_number, on_open_callback)
            self.__confirm_delivery = False
            self.__ack_nack_callback_func = None
            
        @property
        def has_confirm_delivery(self):
            return self.__confirm_delivery
        
        def set_confirm_delivery(self, wait_select_ok=True):
            if wait_select_ok:
                def callback_select_ok(method_frame):
                    if method_frame.method.NAME == 'Confirm.SelectOk':
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"[Channel {self.channel_number}] Confirm delivery mode is set")
                        return True
                    else:
                        raise TechnicalException(f"[Channel {self.channel_number}] Unexpected method frame {method_frame} (expected Confirm.SelectOk)")
                wait_declare_ok = WaitIsTrueOnEvent(f"wait confirm select is OK", callback_select_ok)
                
                self.confirm_delivery(self.__ack_nack_callback, wait_declare_ok.on_event)
                
                wait_declare_ok.execute()
            else:
                self.confirm_delivery(self.__ack_nack_callback)
                
            self.__confirm_delivery = True
            
        def set_ack_nack_callback(self, func):
            self.__ack_nack_callback_func = func
            
        def __ack_nack_callback(self, method_frame:pika.frame.Method):
            if self.__ack_nack_callback_func:
                self.__ack_nack_callback_func(method_frame)
        
        def exchange_declare(self, exchange, exchange_type=ExchangeType.direct, passive=False, durable=False, auto_delete=False, internal=False, arguments=None, wait_declare_ok=True):
            if wait_declare_ok:
                def callback_declare_ok(method_frame):
                    if method_frame.method.NAME == 'Exchange.DeclareOk':
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"[Channel {self.channel_number}] Exchange declare is OK")
                        return True
                    else:
                        raise TechnicalException(f"[Channel {self.channel_number}] Unexpected method frame {method_frame} (expected Exchange.DeclareOk)")
                wait_declare_ok = WaitIsTrueOnEvent(f"wait exchange declare is OK", callback_declare_ok)
                
                super().exchange_declare(exchange, exchange_type=exchange_type, passive=passive, durable=durable, auto_delete=auto_delete, internal=internal, arguments=arguments, callback=wait_declare_ok.on_event)
                
                wait_declare_ok.execute()
            else:
                return super().exchange_declare(exchange, exchange_type=exchange_type, passive=passive, durable=durable, auto_delete=auto_delete, internal=internal, arguments=arguments, callback=None)
        
        def queue_declare(self, queue, passive=False, durable=False, exclusive=False, auto_delete=False, arguments=None, wait_declare_ok=True):
            if wait_declare_ok:
                def callback_declare_ok(method_frame):
                    if method_frame.method.NAME == 'Queue.DeclareOk':
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"[Channel {self.channel_number}] Queue declare is OK")
                        return True
                    else:
                        raise TechnicalException(f"[Channel {self.channel_number}] Unexpected method frame {method_frame} (expected Queue.DeclareOk)")
                wait_declare_ok = WaitIsTrueOnEvent(f"wait queue declare is OK", callback_declare_ok)
                
                super().queue_declare(queue, passive=passive, durable=durable, exclusive=exclusive, auto_delete=auto_delete, arguments=arguments, callback=wait_declare_ok.on_event)
                
                wait_declare_ok.execute()
            else:
                return super().queue_declare(queue, passive=passive, durable=durable, exclusive=exclusive, auto_delete=auto_delete, arguments=arguments, callback=None)
        
        def queue_bind(self, queue, exchange, routing_key=None, arguments=None, wait_bind_ok=True):
            if wait_bind_ok:
                def callback_bind_ok(method_frame):
                    if method_frame.method.NAME == 'Queue.BindOk':
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"[Channel {self.channel_number}] Bind is OK")
                        return True
                    else:
                        raise TechnicalException(f"[Channel {self.channel_number}] Unexpected method frame {method_frame} (expected Queue.BindOk)")
                wait_bind_ok = WaitIsTrueOnEvent(f"wait bind is OK", callback_bind_ok)
                
                super().queue_bind(queue, exchange, routing_key=routing_key, arguments=arguments, callback=wait_bind_ok.on_event)
                
                wait_bind_ok.execute()
            else:
                return super().queue_bind(queue, exchange, routing_key=routing_key, arguments=arguments)


    class RMQSelectConnection(pika.adapters.select_connection.SelectConnection):
        """
        Add features to pika.adapters.select_connection.SelectConnection:
          - Be able to wait until connection is opened.
          - Be able to wait until channel is opened.
          - Use default on_xxx methods to raise on unexpected behaviour
        Note: start_ioloop is automatically started if wait_is_opened is True
        """
        def __init__(self, parameters=None,
                     on_open_callback=None,
                     on_open_error_callback=None,
                     on_close_callback=None,
                     custom_ioloop=None,
                     internal_connection_workflow=True, 
                     wait_is_opened=True):
            """
            Note: Parameters on_open_callback and on_open_error_callback are omitted if waiting until connection is opened.
            """
            # Declare first private variables in case of open error while closing
            self.__is_closing = False
            self.__on_close_callback = on_close_callback
            self.__ioloop_started = False
            self.__func_ioloop_start = None
            self.__channels = []
            
            if wait_is_opened:
                class WaitConnection(Redo):
                    def __init__(self):
                        super().__init__(f"redo connection is opened")
                        self.__connection = None
                        self.__exception = None
                        
                    def _process(self):
                        return (self.__connection, self.__exception)
                    
                    def on_open(self, connection):
                        self.__connection = connection
                    
                    def on_open_error(self, connection, exception):
                        self.__connection = connection
                        self.__exception = exception
                wait_connection = WaitConnection()
                wait_connection.redo_while( (None,None) )
                
                super().__init__(parameters=parameters,
                                 on_open_callback=wait_connection.on_open,
                                 on_open_error_callback=wait_connection.on_open_error,
                                 on_close_callback=self.__on_close,
                                 custom_ioloop=custom_ioloop,
                                 internal_connection_workflow=internal_connection_workflow)
                
                self.start_ioloop()
                conn, exc = wait_connection.execute()
                if exc is not None:
                    raise exc
                if conn is not self:
                    raise TechnicalException(f"Unexpected connection object after waiting open (obtained: {id(conn)} ; expected: {id(self)})")
            
            else:
                super().__init__(parameters=parameters,
                                 on_open_callback=on_open_callback,
                                 on_open_error_callback=on_open_error_callback,
                                 on_close_callback=self.__on_close,
                                 custom_ioloop=custom_ioloop,
                                 internal_connection_workflow=internal_connection_workflow)
        
        def __on_close(self, connection, exception):
            if not self.__is_closing and exception is not None:
                if isinstance(exception, ConnectionClosedByClient) and exception.reply_code != 200:
                    raise exception
            
            if self.__on_close_callback:
                self.__on_close_callback(connection, exception)
        
        def start_ioloop(self, in_thread=True, raise_if_already_started=True):
            if self.__ioloop_started and raise_if_already_started:
                raise TechnicalException(f"IO loop is already started")
            
            if in_thread:
                if self.__func_ioloop_start is not None and self.__func_ioloop_start.is_alive():
                    raise TechnicalException(f"IO loop is running in a thread")
                
                if self.__func_ioloop_start is None:
                    name = f"[Connection {id(self)}] start ioloop"
                    self.__func_ioloop_start = FunctionThreaded(self.start_ioloop, kwargs={'in_thread':False}, name=name)
                self.__func_ioloop_start.start()
            else:
                self.__ioloop_started = True
                self.ioloop.start()
                self.__ioloop_started = False
                
        def stop_ioloop(self):
            if self.__ioloop_started:
                # self.ioloop.stop()
                self.ioloop.add_callback_threadsafe(self.ioloop.stop)
                
        
        def channel(self, channel_number=None, on_open_callback=None, wait_is_opened=True):
            """
            Override channel method to enable waiting until channel is opened.
            Note: Parameter on_open_callback is omitted if waiting until channel is opened.
            """
            res = None
            
            if wait_is_opened:
                # Create wait redo
                class WaitChannel(Redo):
                    def __init__(self):
                        super().__init__(f"redo channel is opened")
                        self.__channel = None
                        
                    def _process(self):
                        return self.__channel
                    
                    def set_channel(self, channel):
                        self.__channel = channel
                        
                wait_channel = WaitChannel()
                wait_channel.with_timeout(3)
                wait_channel.redo_while_none()
                
                # Make maxium 3 retries
                nb_retries = 3
                for n in range(nb_retries):
                    # Call channel method
                    res = super().channel(channel_number=channel_number, on_open_callback=wait_channel.set_channel)
                    
                    try:
                        res = wait_channel.execute()
                    except TimeoutException as exc:
                        if n + 1 == nb_retries:
                            raise exc
                        else:
                            if Tools.do_log(logger, logging.DEBUG):
                                logger.debug("Channel was not open in time, retry")
                        continue
                    else:
                        break
            else:
                res = super().channel(channel_number=channel_number, on_open_callback=on_open_callback)
            
            if res is None:
                raise TechnicalException(f"Failed to get channel")
            self.__channels.append(res)
            
            return res

        def _create_channel(self, channel_number, on_open_callback):
            """Override _create_channel method to use RMQChannel instead of pika.channel.Channel"""
            return RMQChannel(self, channel_number, on_open_callback)


class RMQSelectClient(RMQClient):
    def __init__(self):
        super().__init__('RMQSelectClient')
    
    def _new_connection_parameters(self):
        connection_kwargs = self._connection_kwargs
        wait_connection = connection_kwargs.pop('wait_connection', True)
        return pika.ConnectionParameters(**connection_kwargs), wait_connection
    
    def _new_connection(self):
        if self.rapid_close:
            raise TechnicalException(f"Rapid close is not managed for select connection")
        connection_parameters, wait_connection = self._new_connection_parameters()
        return RMQSelectConnection(parameters=connection_parameters, wait_is_opened=wait_connection)
    
    def _close_connection(self, connection=None):
        if connection is None:
            connection = self.connection
        connection.stop_ioloop()
        super().stop_consuming()
    
    def _new_publisher(self, connection, channel, exchange, routing_key, nb_runners):
        res = RMQSelectPublisher(self, connection, channel, exchange, routing_key, nb_runners)
        return res
        

class RMQSelectPublisher(RMQPublisher):
    """
    Implementation of a publisher for long and intensive publishes.
    """
    
    def __init__(self, client, connection, channel, exchange, routing_key, nb_runners=None):
        super().__init__(client, connection, channel, exchange, routing_key)
        
        # Manage message publish
        self.__publish_lock = threading.Lock()
        self.__publish_nb_runners = nb_runners if nb_runners is not None else 100
        self.__publish_running_nb = 0
        self.__publish_queue = queue.Queue(maxsize=self.__publish_nb_runners * 2)
        self.__publish_counter = 0
        self.__published_counter = 0
        
        # Manage ack/nack
        self.__ack_counter = 0
        self.__nack_counter = 0
        self.__delivery_tag_counter = 0
        self.__ack_nack_lock = threading.Lock()
        # self.__ack_nack_queue = queue.Queue(maxsize=self.__publish_nb_runners * 2)
        self.__ack_nack_queue = queue.Queue()
        self.__last_log_index = 0
        
        # Set ack/nack callback
        channel.set_ack_nack_callback(self.__ack_nack_callback)

    def flush(self):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] flushing...")
        self.__publish_queue.join()
        self.__ack_nack_queue.join()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] flushed")

    def __publish_start(self):
        with self.__publish_lock:
            if self.__publish_running_nb < self.__publish_nb_runners:
                self.__publish_running_nb += 1
            else:
                return
        
        self.__publish_next()

    def __publish_next(self):
        try:
            msg_item = self.__publish_queue.get(block=False)
        except queue.Empty:
            with self.__publish_lock:
                self.__publish_running_nb -= 1
            return
        
        self.client.connection.ioloop.add_callback_threadsafe(
            lambda: self.__publish_message(msg_item[0], **msg_item[1]) )
        
        self.__publish_queue.task_done()

    def __publish_message(self, body, **kwargs):
        # Add delivery tag in queue
        with self.__ack_nack_lock:
            self.__delivery_tag_counter += 1
            del_tag = self.__delivery_tag_counter
        self.__ack_nack_queue.put(del_tag)
        
        # Publish message
        # RMQPublisher.publish(self, body, **kwargs)
        self.client.connection.ioloop.add_callback_threadsafe(
            lambda: RMQPublisher.publish(self, body, **kwargs) )
        self.__published_counter += 1

        # Manage to publish next message
        self.__publish_next()

    def publish(self, body, **kwargs):
        # Add message to publish in queue
        self.__publish_queue.put((body, kwargs))
        self.__publish_counter += 1

        # Start publishing if it is not already running
        self.__publish_start()
        
    def __ack_nack_callback(self, method_frame:pika.frame.Method):
        if method_frame.method.NAME == 'Basic.Ack':
            self.__ack_counter += 1
            del_tag_obtained = method_frame.method.delivery_tag
            del_tag_expected = self.__get_next_delivery_tag_from_ack_nack_queue()
            if method_frame.method.multiple:
                while del_tag_expected < del_tag_obtained:
                    del_tag_expected = self.__get_next_delivery_tag_from_ack_nack_queue()
                if del_tag_expected > del_tag_obtained:
                    raise TechnicalException(f"[{self.name}] Ack: inconsistent multiple delivery tags (obtained: {del_tag_obtained} ; expected: {del_tag_expected})")
            else:
                if del_tag_obtained != del_tag_expected:
                    raise TechnicalException(f"[{self.name}] Ack: inconsistent delivery tags (obtained: {del_tag_obtained} ; expected: {del_tag_expected})")
                
        elif method_frame.method.NAME == 'Basic.Nack':
            self.__nack_counter += 1
            del_tag_obtained = method_frame.method.delivery_tag
            del_tag_expected = self.__get_next_delivery_tag_from_ack_nack_queue()
            if del_tag_obtained != del_tag_expected:
                raise TechnicalException(f"[{self.name}] Nack: inconsistent delivery tags (obtained: {del_tag_obtained} ; expected: {del_tag_expected})")
            
        else:
            raise TechnicalException(f"Unexpected method frame {method_frame}")
        # if self.__delivery_tag_counter // 1000 != self.__last_log_index:
        #     logger.warning(f"++++++++++++++++++++++++++++++++++++++++ __ack_nack_callback: publish queue size: {self.__publish_queue.qsize()} ; delivery tag counter: {self.__delivery_tag_counter} ; ack/nack queue size: {self.__ack_nack_queue.qsize()}")
        #     self.__last_log_index = self.__delivery_tag_counter // 1000
            
    def __get_next_delivery_tag_from_ack_nack_queue(self):
        res = self.__ack_nack_queue.get(block=False)
        self.__ack_nack_queue.task_done()
        return res
    
    def _log_stats(self, level=logging.DEBUG):
        if Tools.do_log(logger, level):
            logger.log(level, f"[{self.name}] delivery tag: {self.__delivery_tag_counter} ; nack: {self.__nack_counter}")
        









#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import os
from holado_system.system.command.command import Command
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.path_manager import PathManager

logger = logging.getLogger(__name__)

class ProtobufCompiler(object):
    def __init__(self):
        self.__protoc_exe_path = "protoc"
        self.__proto_dest = []
        
        # If available, use protoc packaged in grpc_tools, in order to avoid a protoc installation and to use same protoc version than GRpcCompiler  
        from holado_grpc.ipc.rpc.grpc_compiler import GRpcCompiler
        if GRpcCompiler.is_available():
            self.__protoc_exe_path = "python -m grpc_tools.protoc"
    
    @property
    def protoc_exe_path(self):
        return self.__protoc_exe_path
    
    @protoc_exe_path.setter
    def protoc_exe_path(self, path):
        self.__protoc_exe_path = path
    
    def register_proto_path(self, proto_path, destination_path, file_path = None):
        """
        Register a triplet of {proto path, destination folder path, proto file path}.
        If file_path is None, all files/directories of proto_path are explored.
        """
        self.__proto_dest.append((proto_path, destination_path, file_path))
    
    def compile_all_proto(self):
        """Compile all registered proto recursively from proto paths to each associated destination folder paths."""
        for proto_path, destination_path, file_path in self.__proto_dest:
            if file_path is None:
                if os.path.isfile(proto_path):
                    proto_file_path = proto_path
                    proto_path = os.path.dirname(proto_file_path)
                    self.__compile_proto(proto_path, destination_path, proto_file_path)
                elif os.path.isdir(proto_path):
                    self.__compile_all_proto(proto_path, destination_path, proto_path)
                else:
                    raise TechnicalException(f"Unmanaged path '{proto_path}'")
            else:
                self.__compile_all_proto(proto_path, destination_path, file_path)
            
    def __compile_all_proto(self, proto_path, destination_path, current_proto_path, do_create_init_py=False):
        if os.path.isfile(current_proto_path):
            self.__compile_proto(proto_path, destination_path, current_proto_path)
        elif os.path.isdir(current_proto_path):
            lp = os.listdir(current_proto_path)
            for cp in lp:
                cur_proto_path = os.path.join(current_proto_path, cp)
                self.__compile_all_proto(proto_path, destination_path, cur_proto_path)
            if do_create_init_py:
                self.__create_init_py(proto_path, destination_path, current_proto_path)
        else:
            raise TechnicalException(f"Unmanaged path '{current_proto_path}'")
    
    def __create_init_py(self, proto_path, destination_path, proto_file_path):
        """Create __init__.py file in destination folder."""
        if not os.path.isdir(proto_file_path):
            raise TechnicalException(f"Proto path '{proto_file_path}' is not a directory")
        
        relative_proto_path = proto_file_path.replace(proto_path,'').strip('/')
        dest_proto_path = os.path.join(destination_path, relative_proto_path)
        init_path = os.path.join(dest_proto_path, '__init__.py')
        
        pm = PathManager()
        pm.makedirs(dest_proto_path, is_directory=True)
        
        open(init_path, "wt")
        
        logger.info(f"Created __init__.py in destination directory '{relative_proto_path}'")
    
    def __compile_proto(self, proto_path, destination_path, proto_file_path):
        """Compile proto file into destination folder path."""
        if not os.path.isfile(proto_file_path):
            raise TechnicalException(f"Proto path '{proto_file_path}' is not a file")
        
        ext = os.path.splitext(proto_file_path)[1]
        if ext != ".proto":
            return
        
        pm = PathManager()
        pm.makedirs(destination_path, is_directory=True)
        
        cmd = self._build_compile_proto_command(proto_path, destination_path, proto_file_path)
        command = Command(cmd, do_raise_on_stderr=True)
        command.start()
        command.join()
        
        if command.error is not None:
            raise FunctionalException(f"Error while compiling proto file '{proto_file_path}' from '{proto_path}' into '{destination_path}': {command.error}")
        logger.info(f"Compiled proto '{proto_file_path.replace(proto_path,'')}'")

    def _build_compile_proto_command(self, proto_path, destination_path, proto_file_path):
        if os.path.exists(self.protoc_exe_path) and " " in self.protoc_exe_path:
            return f'"{self.protoc_exe_path}" --proto_path="{proto_path}" --python_out="{destination_path}" "{proto_file_path}"'
        else:
            return f'{self.protoc_exe_path} --proto_path="{proto_path}" --python_out="{destination_path}" "{proto_file_path}"'
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc

logger = logging.getLogger(__name__)


class GetOrCreateObject(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        self.__name = name
        
    @property
    def name(self):
        return self.__name
    
    def has(self, name):
        return self._goc_has_object(name)
    
    def get(self, name):
        if self.has(name):
            return self._goc_get_object(name)
        else:
            return None
    
    def create(self, name, *args, **kwargs):
        res = self._goc_new_object(name, *args, **kwargs)
        self._goc_set_object(name, res)
        return res
    
    def get_or_create(self, name, *args, **kwargs):
        if self.has(name):
            return self.get(name)
        else:
            return self.create(name, *args, **kwargs)

    def _goc_has_object(self, name):
        raise NotImplementedError()

    def _goc_get_object(self, name):
        raise NotImplementedError()

    def _goc_new_object(self, name, *args, **kwargs):
        raise NotImplementedError()

    def _goc_set_object(self, name, obj):
        raise NotImplementedError()


class GetOrCreateVariable(GetOrCreateObject):
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name):
        super().__init__(name)
        self.__func_variable_manager = None
        
    def initialize(self, func_variable_manager):
        self.__func_variable_manager = func_variable_manager
        
    @property
    def __variable_manager(self):
        return self.__func_variable_manager()
        
    def _goc_has_object(self, name):
        return self.__variable_manager.exists_variable(name)
        
    def _goc_get_object(self, name):
        return self.__variable_manager.get_variable_value(name)

    def _goc_set_object(self, name, obj):
        self.__variable_manager.register_variable(name, obj)
        

class GetOrCreateVariableObject(GetOrCreateVariable):
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name):
        super().__init__(name)
        
    def initialize(self, func_variable_manager):
        super().initialize(func_variable_manager)
        
    def get(self, name):
        varname = self._goc_get_varname(name)
        return varname, super().get(name)
    
    def create(self, name, *args, **kwargs):
        varname = self._goc_get_varname(name)
        return varname, super().create(name, *args, **kwargs)
    
    def _goc_has_object(self, name):
        varname = self._goc_get_varname(name)
        return super()._goc_has_object(varname)
        
    def _goc_get_object(self, name):
        varname = self._goc_get_varname(name)
        return super()._goc_get_object(varname)

    def _goc_set_object(self, name, obj):
        varname = self._goc_get_varname(name)
        super()._goc_set_object(varname, obj)
        
    def _goc_get_varname(self, name):
        return f"GOC_{self.name.upper()}_{str(name).replace(' ','').upper()}"
        
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado.common.context.context import Context
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_ui.ui.exceptions.focus_driver_exception import FocusDriverException
from holado_ui_selenium.ui.gui.selenium.handlers.enums import BrowserTypes
from holado_ui_selenium.ui.gui.selenium.drivers.selenium_driver import SeleniumDriver

logger = logging.getLogger(__name__)



class SeleniumUIManager(Context):
    """ Tools to manage existing web browsers.
    It adds features to UIManager.
    """
    
    def __init__(self, ui_manager, path_manager):
        super().__init__("SeleniumUIManager")
        
        self.__ui_manager = ui_manager
        self.__path_manager = path_manager
    
    @property
    def current_browser(self):
        """
        @return Current browser instance
        """
        # Create browser if needed
        if not self.has_current_browser():
            #TODO: add default browser type in a config file
            browser_type = BrowserTypes.Chrome
            
            # If not already set, set browser with configured version
            if not self.has_browser(browser_type):
                self.set_browser(browser_type)
            
            # Update current browser ID
            self.set_current_browser_id(self.get_browser_id(browser_type))
        
        # When last page of a browser is closed, the browser is removed from registered driver, since it is not visible anymore.
        # Thus, if current browser is not anymore in registered drivers, push it again
        browser_id = self.current_browser_id
        if not self.__ui_manager.has_driver(browser_id):
            self.__ui_manager.push_driver(browser_id, self.get_browser_of_id(browser_id), False)
        
        # Verify that active driver is expected browser
        try:
            self.__ui_manager.verify_driver_is_in_focus(self.get_current_browser_id())
        except FocusDriverException as exc:
            raise FocusDriverException(f"Current active UI is not expected browser (usually it is solved by switching into the browser before): expected: '{self.get_current_browser_id()}' ; current active: '{self.__ui_manager.in_focus_driver_info.uid}'") from exc
        
        return self.__ui_manager.in_focus_driver_info.driver
    
    @property
    def current_browser_id(self):
        """
        @return Current browser ID
        """
        return self.get_object("current_browser_id")

    @current_browser_id.setter
    def current_browser_id(self, browser_id):
        """
        Set current browser ID
        @param browser_id Browser ID
        """
        self.set_object("current_browser_id", browser_id, False)

    @property
    def has_current_browser(self):
        """
        @return True if context contains current browser instance
        """
        return self.has_object("current_browser_id")

    def set_current_browser_type(self, browser_type):
        """
        Set current browser type
        @param browser_type Browser type
        """
        self.current_browser_id = self.get_browser_id(browser_type)

    def get_browser(self, browser_type):
        """
        @param browser_type Browser type
        @return Browser instance of given type
        """
        if not self.has_browser(browser_type):
            self.set_browser(browser_type)
        return self.get_object(self.get_browser_id(browser_type))

    def get_browser_of_id(self, browser_id):
        """
        @param browser_id Browser ID
        @return Browser instance of given ID
        """
        return self.get_object(browser_id)

    def get_browser_id(self, browser_type):
        """
        @param browser_type Browser type
        @return Browser ID of given type
        """
        return f"{browser_type.name}_browser"

    def has_browser(self, browser_type):
        """
        @param browser_type Browser type
        @return True if context contains browser instance of given type
        """
        return self.has_object(self.get_browser_id(browser_type))

    def set_browser(self, browser_type):
        """
        @param browser_type Browser type
        """
        if self.has_Browser(browser_type):
            logger.info(f"setting a new browser of type '{browser_type.name}'")
            
            uid = self.get_browser_id(browser_type)
            
            # Create and initialize new browser
            selenium_driver = SeleniumDriver(browser_type, self.__path_manager)
            self.__ui_manager.initialize_ui_driver(uid, selenium_driver)
            
            # Add driver in UI manager
            self.__ui_manager.push_driver(self.get_browser_id(browser_type), self.get_browser(browser_type), False)
        else:
            raise TechnicalException(f"Browser of type '{browser_type.name}' is already set")

    def set_browser_lifetime_to(self, lifetime_context, browser_type):
        """
        Set browser lifetime to given context
        @param lifetime_context Lifetime context
        @param browser_type Browser type
        """
        if self.has_browser(browser_type):
            browser_id = self.get_browser_id(browser_type)
            self.__ui_manager.set_driver_lifetime_to(lifetime_context, browser_id)
        else:
            raise TechnicalException(f"No browser of type '{browser_type.name}' is set")

    def switch_to_browser(self, browser_type):
        """
        Make given browser type as active UI driver
        @param browser_type Browser type
        """
        # Set current browser type
        self.set_current_browser_type(browser_type)
        
        browser_id = self.get_browser_id(browser_type)
        if not self.__ui_manager.has_driver(browser_id):
            if not self.has_browser(browser_type):
                self.set_browser(browser_type)
            else:
                self.__ui_manager.push_driver(browser_id, self.get_browser_of_id(browser_id), False)
        else:
            self.__ui_manager.switch_to_driver(browser_id)
        
        # If browser has already an opened window, focus on it
        if self.current_browser.is_opened:
            handle = self.current_browser.internal_driver.get_window_handle()
            
            # Use an alert to force focus
            self.current_browser.internal_driver.execute_script("alert('Test')")
            self.current_browser.internal_driver.swith_to.alert.accept()
            
            self.current_browser.internal_driver.swith_to.window(handle)

    def switch_to_browser(self, browser_id):
        """
        Make browser of given ID as active UI driver
        @param browser_id ID of already opened browser
        """
        # Repush browser
        self.__ui_manager.switch_to_driver(browser_id)
        
        # If browser has already an opened window, focus on it
        if self.current_browser.is_opened:
            handle = self.current_browser.internal_driver.get_window_handle()
            self.current_browser.internal_driver.switchTo().window(handle)
    
    def close_browser(self, browser_type, lifetime_context=None):
        """
        Close given browser type
        @param browser_type Browser type
        @param lifetime_context Lifetime context
        @return True if a driver was closed
        """
        res = False
        
        if self.has_browser(browser_type) and self.get_browser(browser_type).is_opened:
            # Close browser
            browser_id = self.get_browser_id(browser_type)
            res = self.__ui_manager.close_driver(browser_id, lifetime_context)
        
        return res
    
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.handlers.object import DeleteableObject
from holado_system.system.filesystem.file import File
import json
from holado.common.context.session_context import SessionContext


def __get_path_manager():
    return SessionContext.instance().path_manager


class StreamJSONFile(DeleteableObject):
    """
    Manage a stream JSON file.
    
    A stream JSON file is a file with many root elements that are valid JSON.
    It is useful for example to process workflows of data formatted in JSON.
    An example in HolAdo is the mechanism saving docker container logs in a JSON file, 
    which adds in a single file a log by line in JSON format.
    
    Note: Current implementation assumes that file has a line per JSON object
    """

    def __init__(self, path, auto_flush=True, do_open=False, **open_kwargs):
        super().__init__(f"stream JSON file '{path}'")

        self.__file = File(path, auto_flush=auto_flush, do_open=do_open, **open_kwargs)
        self.__current_line = None
        self.__current_pos = None

    def __enter__(self):
        self.__file.__enter__()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__file.__exit__(exc_type, exc_val, exc_tb)
    
    @property
    def path(self):
        return self.__file.path

    @property
    def internal_file(self):
        return self.__file
    
    @property
    def is_open(self):
        return self.__file.is_open

    def open(self, **kwargs):
        self.__file.open(**kwargs)

    def close(self):
        self.__file.close()

    def read_element_as_json_object(self, **loads_kwargs):
        line = self.__file.readline()
        if not line or len(line.strip()) == 0:
            return None
        
        return json.loads(line, **loads_kwargs)
    
    def read_elements_as_json_object_list(self, **loads_kwargs):
        res = []
        while True:
            element = self.read_element_as_json_object(**loads_kwargs)
            if element is not None:
                res.append(element)
            else:
                break
        return res
    
    def write_element_json_object(self, data, **dumps_kwargs):
        dumps_kwargs['indent'] = None
        
        content = json.dumps(data, **dumps_kwargs)
        self.__file.writelines([content])
    
    def write_elements_json_object_list(self, data_list, **dumps_kwargs):
        for data in data_list:
            self.write_element_json_object(data, **dumps_kwargs)


    @classmethod
    def read_stream_json_file_content_as_json_object_list(cls, path, open_kwargs=None, loads_kwargs=None):
        open_kwargs = open_kwargs if open_kwargs is not None else {}
        loads_kwargs = loads_kwargs if loads_kwargs is not None else {}
        
        file = StreamJSONFile(path, **open_kwargs)
        return file.read_elements_as_json_object_list(**loads_kwargs)

    @classmethod
    def write_stream_xml_file_with_dict_list(cls, path, data_list, open_kwargs=None, dumps_kwargs=None):
        open_kwargs = open_kwargs if open_kwargs is not None else {}
        dumps_kwargs = dumps_kwargs if dumps_kwargs is not None else {}
        
        __get_path_manager().makedirs(path)
        
        file = StreamJSONFile(path, **open_kwargs)
        file.write_elements_json_object_list(data_list, **dumps_kwargs)
        return file







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_ui.ui.handlers.ui_context import UIContext
from holado.common.handlers.enums import AutoNumber
from holado_core.common.exceptions.technical_exception import TechnicalException


class InformationTypes(AutoNumber):
    """
    Information type
    """
    
    Id = ()
    Name = ()
    Label = ()
    

class ZoneGUIContext(UIContext):
    """ Zone context.
    """
    
    def __init__(self, previous_context, information_type:InformationTypes, information):
        super().__init__(previous_context)
        self.__information_type = information_type
        self.__information = information
    
    @property
    def information_type(self) -> InformationTypes:
        """
        @return Information type
        """
        return self.__information_type
    
    @property
    def information(self):
        """
        @return Information
        """
        return self.__information
    
    def get_description(self):
        """
        @return Context description
        """
        res = "zone "
        
        if self.information_type == InformationTypes.Id:
            res += f"of id '{self.information}'"
        elif self.information_type == InformationTypes.Label:
            res += f"'{self.information}'"
        elif self.information_type == InformationTypes.Name:
            res += f"named '{self.information}'"
        else:
            raise TechnicalException(f"Unmanaged information type '{self.information_type.name}'")
        
        return res
    
    @classmethod
    def main_zone(cls):
        """
        @return Context for main zone
        """
        return cls(None, InformationTypes.Name, "main")
        
    @classmethod
    def zones_by_labels(cls, *labels):
        """
        @param labels Labels
        @return Context for successive zones with given labels
        """
        res = None
        for label in labels:
            res = cls(res, InformationTypes.Label, label)
        return res
    
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.typing import Typing
from holado_core.common.exceptions.functional_exception import FunctionalException

logger = logging.getLogger(__name__)


def create_name_value_table_from_json(json_value, recursive=False, uncollapse_list=False):
    from holado_json.ipc.json_converter import JsonConverter
    
    res = TableWithHeader()
    res.header.add_cells_from_contents(["Name", "Value"])
    
    converter = JsonConverter()
    json_value = converter.to_json(json_value)
    
    name_values = convert_json_value_2_name_values(json_value, recursive=recursive, uncollapse_list=uncollapse_list)
    for name, value in name_values:
        res.add_row(cells_content=(name, value))
    
    return res
    
def create_table_with_header_from_json(json_value, recursive=False, uncollapse_list=False):
    from holado_json.ipc.json_converter import JsonConverter
    
    res = TableWithHeader()
    
    converter = JsonConverter()
    json_value = converter.to_json(json_value)
    
    if isinstance(json_value, list):
        json_list = json_value
    else:
        json_list = [json_value]
    
    for json_val in json_list:
        json_val = converter.to_json(json_val)
        
        name_values = convert_json_value_2_name_values(json_val, recursive=recursive, uncollapse_list=uncollapse_list)
        
        # Add new columns if needed
        for name, value in name_values:
            if not res.has_column(name, raise_exception=False):
                res.add_column(name=name, cells_content = [None]*res.nb_rows)
        
        # Fill table
        values_by_name = {name: value for name, value in name_values}
        res.add_row(contents_by_colname = values_by_name)
        
    # Order columns
    if res.nb_columns > 0:
        res.order_columns(names = sorted(res.get_column_names()))
        
    return res
    
def convert_json_value_2_name_values(json_value, recursive=False, uncollapse_list=False, as_dict=False, sorted_=False):
    """Convert json object as a list of (name, value).
    If as_dict is True, the result is a dict of {name:value}
    """
    res = []
    __fill_name_values_with_json_value(res, None, json_value, recursive=recursive, uncollapse_list=uncollapse_list)
    
    if sorted_:
        res = sorted(res)
    if as_dict:
        res = {e[0]:e[1] for e in res}
    return res
    
def __fill_name_values_with_json_value(res, name_prefix, json_value, recursive=False, uncollapse_list=False):
    if isinstance(json_value, dict):
        __fill_name_values_with_json_dict(res, name_prefix, json_value, recursive=recursive, uncollapse_list=uncollapse_list)
    elif isinstance(json_value, list):
        __fill_name_values_with_json_list(res, name_prefix, json_value, recursive=recursive, uncollapse_list=uncollapse_list)
    else:
        res.append((name_prefix, json_value))
    
def __fill_name_values_with_json_dict(res, name_prefix, json_value, recursive=False, uncollapse_list=False):
    if not isinstance(json_value, dict):
        raise TechnicalException(f"json value is expected to be a dict (obtained: {Typing.get_object_class_fullname(json_value)}")
    
    sorted_dict = dict(sorted(json_value.items()))
    for name, value in sorted_dict.items():
        new_prefix = f"{name_prefix}.{name}" if name_prefix is not None and len(name_prefix) > 0 else name
        if recursive:
            __fill_name_values_with_json_value(res, new_prefix, value, recursive=recursive, uncollapse_list=uncollapse_list)
        else:
            res.append((new_prefix, value))
    
def __fill_name_values_with_json_list(res, name_prefix, json_value, recursive=False, uncollapse_list=False):
    if not isinstance(json_value, list):
        raise TechnicalException(f"json value is expected to be a list (obtained: {Typing.get_object_class_fullname(json_value)}")
    
    if not uncollapse_list:
        res.append((name_prefix, json_value))
        return
        
    if len(json_value) == 0:
        res.append((name_prefix+"[]", None))
        return
        
    for index, value in enumerate(json_value):
        new_prefix = f"{name_prefix}[{index}]" if name_prefix is not None else f"[{index}]"
        __fill_name_values_with_json_value(res, new_prefix, value, recursive=recursive, uncollapse_list=uncollapse_list)


def set_object_attributes_with_json_dict(obj, json_value):
    if not isinstance(json_value, dict):
        raise TechnicalException(f"json value is expected to be a dict (obtained: {Typing.get_object_class_fullname(json_value)}")
    
    for attr_name, attr_value_json in json_value.items():
        if hasattr(obj, attr_name):
            if isinstance(attr_value_json, dict) and not isinstance(getattr(obj, attr_name), dict):
                set_object_attributes_with_json_dict(getattr(obj, attr_name), attr_value_json)
            else:
                setattr(obj, attr_name, attr_value_json)
        else:
            raise FunctionalException(f"Object doesn't have attribute '{attr_name}' (object type: {Typing.get_object_class_fullname(obj)})")


def update_json_object_with_missing_attributes(json_obj, json_ref, recursive=False):
    if not isinstance(json_obj, dict):
        raise TechnicalException(f"json object is expected to be a dict (obtained: {Typing.get_object_class_fullname(json_obj)}")
    if not isinstance(json_ref, dict):
        raise TechnicalException(f"json reference is expected to be a dict (obtained: {Typing.get_object_class_fullname(json_ref)}")

    for key, value in json_ref.items():
        if key not in json_obj:
            json_obj[key] = value
        elif recursive and isinstance(json_obj[key], dict) and isinstance(value, dict):
            update_json_object_with_missing_attributes(json_obj[key], value, recursive=recursive)
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.inspectors.inspector import Inspector
from holado_core.common.finders.tools.enums import FindType
from holado_core.common.finders.then_finder import ThenFinder
import abc
from holado_core.common.inspectors.tools.inspect_context import InspectContext
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters

logger = logging.getLogger(__name__)



class ElementInspector(Inspector):
    """ Base class for inspector on elements.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        super().__init__(module_name)
    
    def get_finder_element_in(self, finder_container, finder_element, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Find element with given finder, in given container
        @param finder_container Finder for container
        @param finder_element Finder for element to find
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        if finder_container is None or finder_element is None:
            return self._build_result_finder(finder_element, inspect_context, inspect_parameters)

        res = ThenFinder(f"{finder_element.element_description} in {finder_container.element_description}")

        # Find container
        res.set_next_finder(finder_container)

        # Then find element in container
        res.set_next_finder(FindType.In, True, finder_element)

        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
import logging
import os.path
from holado.common.context.context import Context
from holado_core.common.tools.tools import Tools
import threading
from holado.common.handlers.enums import ObjectStates
from holado.holado_config import Config


logger = logging

def initialize_logger():
    global logger
    logger = logging.getLogger(__name__)


class SessionContext(Context):
    TSessionContext = None
    
    # Singleton management
    __instance = None
    __is_resetting_instance = False
    
    @staticmethod
    def instance() -> TSessionContext:
        # Note: If session context istance is under reset, consider it is already None
        if SessionContext.__is_resetting_instance:
            return None
        
        if SessionContext.__instance is None:
            SessionContext.__instance = SessionContext.TSessionContext()
            logger.debug(f"Created session context of type {SessionContext.TSessionContext}")
            # logging.log(45, f"Created session context of type {SessionContext.TSessionContext}")
            # import traceback
            # logging.log(45, "".join(traceback.format_list(traceback.extract_stack())))
        return SessionContext.__instance
    
    @staticmethod
    def has_instance() -> bool:
        if SessionContext.__is_resetting_instance:
            return False
        else:
            return SessionContext.__instance is not None
    
    @staticmethod
    def _reset_instance():
        if SessionContext.__instance is not None:
            logger.debug(f"Resetting session context")
            SessionContext.__is_resetting_instance = True
            SessionContext.__instance = None
            SessionContext.__is_resetting_instance = False
            logger.debug(f"Reset of session context")
    
    def __init__(self, name="Session"):
        super().__init__(name, with_post_process=True)
        
        self.__with_session_path = None
        
        from holado.common.context.service_manager import ServiceManager
        self.__service_manager = ServiceManager(type(self))
        
        # Manage multitasking
        self.__multitask_lock = threading.RLock()
        self.__multitask_step_lock = threading.RLock()
        
    def __del__(self):
        # Note: Override Object.__del__ since it supposes that SessionContext is initialized, whereas SessionContext is deleting
        try:
            if self.object_state in [ObjectStates.Deleting, ObjectStates.Deleted]:
                return
            
            self.delete_object()
        except Exception as exc:
            if "Python is likely shutting down" in str(exc):
                # Simply return
                return
            else:
                raise exc
    
    def _delete_object(self):
        # if Tools.do_log(logger, logging.DEBUG):
        #     logger.debug("Interrupting and unregistering all threads for scenario [{}]".format(scenario.name))
        if self.has_threads_manager:
            logger.info(f"Delete session context - Interrupting and unregistering all threads...")
            self.threads_manager.interrupt_all_threads(scope="Session")
            self.threads_manager.unregister_all_threads(scope="Session", keep_alive=False)
        
        # Delete session context
        logger.info(f"Delete session context - Deleting context objects...")
        super()._delete_object()
        
    def configure(self, session_kwargs=None):
        """
        Override this method to configure the session context before new session creation.
        It is usually used to register new services.
        """
        if session_kwargs is None:
            session_kwargs = {}
            
        self.__with_session_path = session_kwargs.get("with_session_path", True)
        
        # Create this thread context
        self.multitask_manager.get_thread_context()
    
    def initialize(self, session_kwargs=None):
        """
        Override this method to initialize the session context after its configuration and new session creation.
        """
        pass
        
    @property
    def services(self):
        return self.__service_manager
    
    @property
    def with_session_path(self):
        return self.__with_session_path
    
    def get_object_getter_eval_string(self, obj, raise_not_found=True):
        from holado_python.standard_library.typing import Typing
        
        name = self.get_object_name(obj)
        if name is not None:
            if self.__service_manager.has_service(name):
                return f"{Typing.get_object_class_fullname(self)}.instance().{name}"
            else:
                return f"{Typing.get_object_class_fullname(self)}.instance().get_object('{name}')"
        
        if raise_not_found:
            from holado_core.common.exceptions.element_exception import ElementNotFoundException
            raise ElementNotFoundException(f"[{self.name}] Failed to find object of id {id(obj)}")
        else:
            return None

    def new_session(self, session_kwargs=None):
        from holado_python.common.tools.datetime import DateTime
        
        # Report session
        report_path = None
        if self.with_session_path:
            # Create new report path for this session
            name = "session_{}".format(DateTime.now(tz=Config.report_timezone).strftime("%Y-%m-%d_%H-%M-%S"))  # @UndefinedVariable
            report_path = self.path_manager.get_reports_path(name)
            logger.info(f"Reports location: {report_path}")
            print(f"Reports location: {report_path}")
        self.report_manager.new_session(report_path)
            
        # Logging configuration
        if self.with_session_path and SessionContext.instance().log_manager.in_file:
            log_filename = os.path.join(report_path, "logs", "report.log")
            self.path_manager.makedirs(log_filename)
            SessionContext.instance().log_manager.set_root_log_file(log_filename)
            SessionContext.instance().log_manager.set_config()
    
    def before_all(self, behave_context):
        log_prefix = f"[Before session] "
        
        with self.__multitask_lock:
            self.behave_manager.set_main_context(behave_context)
            self.report_manager.before_all()
        
        # Set variable with session context instance
        self.variable_manager.register_variable("SESSION_CONTEXT", self)
        
        logger.info(f"{log_prefix}Doing previous session post processes if needed...")
        self.do_persisted_post_processes()
        
    def after_all(self):
        log_prefix = f"[After session] "
            
        # Post processes
        logger.info(f"{log_prefix}Post processing...")
        self.do_post_processes()
        
        with self.__multitask_lock:
            self.report_manager.after_all()
            self.behave_manager.clear()

    def before_feature(self, feature):
        from holado_system.system.global_system import GlobalSystem
        from holado_helper.debug.memory.memory_profiler import MemoryProfiler
        from holado_test.common.context.feature_context import FeatureContext
        from holado_test.test_config import TestConfig
        
        log_prefix = f"[Before feature '{feature.name}'] "
        
        with self.__multitask_lock:
            # Logs
            logger.info("="*150)
            logger.info(f"Feature [{feature.name}]")
            
            logger.info(f"{log_prefix}Begin")
            if self.has_feature_context(is_reference=True):
                from holado_core.common.exceptions.technical_exception import TechnicalException
                raise TechnicalException(f"{log_prefix}A feature context is already defined")
            
            GlobalSystem.log_resource_usage(prefix=log_prefix, level=logging.INFO, logger_=logger)
            if TestConfig.profile_memory_in_features and MemoryProfiler.is_tracker_available():
                MemoryProfiler.create_or_reset_tracker_of_objects_summary_changes("features summary")
                # MemoryProfiler.create_or_reset_tracker_of_objects_changes("features objects")
                
            # Feature context
            feature_context = FeatureContext(feature)
            self.__set_feature_context(feature_context)
            
            # Set variable with feature context instance
            self.variable_manager.register_variable("FEATURE_CONTEXT", feature_context)
            
            # Report
            try:
                self.report_manager.before_feature(feature_context, feature)
            except:
                logger.exception(f"{log_prefix}Error while updating report before feature")
            
            logger.info(f"{log_prefix}End")
        
    def after_feature(self, feature):
        from holado_system.system.global_system import GlobalSystem
        from holado_helper.debug.memory.memory_profiler import MemoryProfiler
        from holado_test.test_config import TestConfig
        
        log_prefix = f"[After feature '{feature.name}'] "
        
        with self.__multitask_lock:
            logger.info(f"{log_prefix}Begin")
            if not self.has_feature_context(is_reference=True):
                from holado_core.common.exceptions.technical_exception import TechnicalException
                raise TechnicalException(f"{log_prefix}No feature context is defined")
            
            if TestConfig.profile_memory_in_features and MemoryProfiler.is_tracker_available():
                MemoryProfiler.log_tracker_diff(name="features summary", prefix=log_prefix, level=logging.INFO, logger_=logger)  # @UndefinedVariable
                # MemoryProfiler.log_tracker_diff(name="features objects", prefix="[After feature] ", level=logging.INFO, logger_=logger)  # @UndefinedVariable
            GlobalSystem.log_resource_usage(prefix=log_prefix, level=logging.INFO, logger_=logger)
            
            # End feature context
            self.get_feature_context().end()
            
            # Report
            try:
                self.report_manager.after_feature(feature)
            except:
                logger.exception(f"{log_prefix}Error while updating report after feature")
            
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"{log_prefix}Deleting feature context")
            self.__delete_feature_context()
            
            logger.info(f"{log_prefix}End")
            logger.info(f"Finished feature [{feature.name}]")
        
    def has_feature_context(self, is_reference=None, do_log=False):
        return self.multitask_manager.has_feature_context(is_reference=is_reference, do_log=do_log)
    
    def get_feature_context(self):
        return self.multitask_manager.get_feature_context()
    
    def __set_feature_context(self, feature_context):
        return self.multitask_manager.set_feature_context(feature_context)
    
    def __delete_feature_context(self):
        return self.multitask_manager.delete_feature_context()
    
    def before_scenario(self, scenario):
        from holado_system.system.global_system import GlobalSystem
        from holado_helper.debug.memory.memory_profiler import MemoryProfiler
        from holado_test.common.context.scenario_context import ScenarioContext
        from holado_test.test_config import TestConfig
        
        log_prefix = f"[Before scenario '{scenario.name}'] "
        
        with self.__multitask_lock:
            logger.info("-"*150)
            logger.info(f"Scenario [{scenario.name}]")
            
            logger.info(f"{log_prefix}Begin")
            if not self.has_feature_context(is_reference=True):
                from holado_core.common.exceptions.technical_exception import TechnicalException
                raise TechnicalException(f"{log_prefix}No feature context is defined")
            
            # Create and initialize ScenarioContext
            scenario_context = ScenarioContext(scenario)
            self.get_feature_context().add_scenario(scenario_context)
            
            # Report: create scenario report
            try:
                self.report_manager.before_scenario(scenario_context, scenario)
            except:
                logger.exception(f"{log_prefix}Error while updating report before scenario")
            
            # Set variable with scenario context instance
            # Note: must be after scenario report creation
            self.get_scenario_context().get_variable_manager().register_variable("SCENARIO_CONTEXT", self.get_scenario_context())
            
            # Behave context
            try:
                self.behave_manager.before_scenario()
            except:
                logger.exception(f"{log_prefix}Error while updating behave context before scenario")
            
            GlobalSystem.log_resource_usage(log_prefix, level=logging.INFO, logger_=logger)
            if TestConfig.profile_memory_in_scenarios and MemoryProfiler.is_tracker_available():
                MemoryProfiler.create_or_reset_tracker_of_objects_summary_changes("scenarios summary")
                # MemoryProfiler.create_or_reset_tracker_of_objects_changes("scenarios objects")
        
            logger.info(f"{log_prefix}Doing previous scenario post processes if needed...")
            self.get_scenario_context().do_persisted_post_processes()
            
            logger.info(f"{log_prefix}End")
            logger.info(f"Start scenario [{scenario.name}]")
        
        
    def after_scenario(self, scenario):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        from holado_system.system.global_system import GlobalSystem
        from holado_helper.debug.memory.memory_profiler import MemoryProfiler
        from holado_test.test_config import TestConfig
        
        log_prefix = f"[After scenario '{scenario.name}'] "
        
        with self.__multitask_lock:
            # End scenario
            if not self.has_feature_context(is_reference=True):
                raise TechnicalException(f"{log_prefix}No feature context is defined")
            if not self.has_scenario_context(is_reference=True):
                raise TechnicalException(f"{log_prefix}No scenario context is defined")
            self.get_scenario_context().end()
            
            # Process actions at scenario end
            try:
                self.after_scenario_end(scenario)
            except:
                logger.exception(f"{log_prefix}Error while processing actions after scenario end")
            
            # Process after scenario
            logger.info(f"{log_prefix}Begin")
            # logger.printf"++++++++++++ scenario: {Tools.represent_object(scenario)}")
            
            logger.info(f"{log_prefix}Resource usage:")
            if TestConfig.profile_memory_in_scenarios and MemoryProfiler.is_tracker_available():
                MemoryProfiler.log_tracker_diff(name="scenarios summary", prefix=log_prefix, level=logging.INFO, logger_=logger)  # @UndefinedVariable
                # MemoryProfiler.log_tracker_diff(name="scenarios objects", prefix="[After scenario] ", level=logging.INFO, logger_=logger)  # @UndefinedVariable
            GlobalSystem.log_resource_usage(prefix=log_prefix, level=logging.INFO, logger_=logger)
            
            # Post processes
            logger.info(f"{log_prefix}Post processing...")
            self.get_scenario_context().scope_manager.reset_scope_level()
            self.get_scenario_context().do_post_processes()
            
            # Report
            logger.info(f"{log_prefix}Generating reports...")
            try:
                self.report_manager.after_scenario(scenario)
            except:
                logger.exception(f"{log_prefix}Error while updating report after scenario")
            
            # Delete scenario context
            logger.info(f"{log_prefix}Deleting scenario context...")
            # if Tools.do_log(logger, logging.DEBUG):
            #     logger.debug("Deleting context of scenario [{}]".format(scenario.name))
            self.__delete_scenario_context()
            
            # Remove all threads
            logger.info(f"{log_prefix}Interrupting and unregistering all threads...")
            # if Tools.do_log(logger, logging.DEBUG):
            #     logger.debug("Interrupting and unregistering all threads for scenario [{}]".format(scenario.name))
            #TODO: For case of multiple scenarios launched in parallel, interrupt only threads related to current scenario (scenario launched by this thread)
            self.threads_manager.interrupt_all_threads(scope="Scenario")
            self.threads_manager.unregister_all_threads(scope="Scenario", keep_alive=False)
            
            logger.info(f"{log_prefix}End")
            logger.info(f"Finished scenario [{scenario.name}]")
    
    def after_scenario_end(self, scenario):
        from holado_report.report.report_manager import ReportManager
        
        # Log error on failing scenario
        category_validation, status_validation, step_failed, step_number, scenario_context, step_context = ReportManager.get_current_scenario_status_information(scenario)
        has_failed = status_validation != "Passed"
        if has_failed:
            msg_list = []
            category_str = f" => {category_validation}" if category_validation else ""
            msg_list.append(f"Scenario {status_validation}{category_str}: {ReportManager.format_context_period(scenario_context)} {ReportManager.format_scenario_short_description(scenario)} - {ReportManager.format_step_short_description(step_failed, step_number, step_context=step_context, dt_ref=scenario_context.start_datetime, has_failed=has_failed)}")
            step_error_message = ReportManager.get_step_error_message(step_failed)
            if step_error_message:
                msg_list.append(step_error_message)
            msg = "\n".join(msg_list)
            logger.error(msg)
    
    def has_scenario_context(self, is_reference=None):
        return self.has_feature_context(is_reference=is_reference) and self.get_feature_context().has_scenario
    
    def get_scenario_context(self):
        return self.get_feature_context().current_scenario
    
    def __delete_scenario_context(self):
        self.get_scenario_context().delete_object()
        
    def before_step(self, step):
        log_prefix = f"[Before step '{step}'] "
        
        with self.__multitask_step_lock:
            from holado_test.common.context.step_context import StepContext
            from holado_core.common.exceptions.technical_exception import TechnicalException
            
            if not self.has_feature_context(is_reference=None):
                # Look again but do logs before raising exception
                self.has_feature_context(is_reference=None, do_log=True)
                raise TechnicalException(f"{log_prefix}No feature context is defined (step: {step})")
            if not self.has_scenario_context(is_reference=None):
                raise TechnicalException(f"{log_prefix}No scenario context is defined (step: {step})")
            scenario_context = self.get_scenario_context()
            
            # Manage step context
            step_context = StepContext(step)
            scenario_context.scope_manager.set_step_context(step_context)
            
            # Update scenario context
            step_level = scenario_context.scope_manager.scope_level("steps")
            if step_level == 0:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"{log_prefix}Add step {step_context} in scenario {scenario_context}")
                scenario_context.add_step(step_context)
            
            # Manage scope in define
            if scenario_context.block_manager.is_in_define \
                    and (scenario_context.block_manager.is_in_sub_define         # needed to add in define the "end define" steps of sub-define \
                         or step.name not in ["end for", "end while", "end function"]):     # needed to not add in define the "end define" steps of current define
                from holado_test.behave.behave import format_step
                step_str = format_step(step)
                scenario_context.block_manager.scope_in_define.add_steps(step_str)
                
                # Set step status
                step_context.status = "defined"
            
            # Report
            try:
                self.report_manager.before_step(step_context, step, step_level)
            except:
                logger.exception(f"{log_prefix}Error while updating report before step")
    
    def after_step(self, step, has_started=True):
        """Process after step
        @param step: step instance
        @param has_started: if False, the step is added but without execution. 
            It is usually True and before_step was called before, except for undefined and skipped steps.
        """
        log_prefix = f"[After step '{step}'] "
        
        with self.__multitask_step_lock:
            from holado_test.common.context.step_context import StepContext
            from holado_core.common.exceptions.technical_exception import TechnicalException
            
            if not self.has_feature_context(is_reference=None):
                raise TechnicalException(f"{log_prefix}No feature context is defined (step: {step})")
            if not self.has_scenario_context(is_reference=None):
                raise TechnicalException(f"{log_prefix}No scenario context is defined (step: {step})")
            scenario_context = self.get_scenario_context()
            
            # Manage step context
            if has_started:
                step_context = scenario_context.scope_manager.get_step_context()
                step_context.end()
                if step_context.status is None:
                    step_context.status = step.status.name
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"{log_prefix}Ended step {step_context} in scenario {scenario_context}")
            else:
                # Manage step context
                step_context = StepContext(step, do_start=False)
                scenario_context.scope_manager.set_step_context(step_context)
                
                # Update scenario context
                step_level = scenario_context.scope_manager.scope_level("steps")
                if step_level == 0:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug(f"{log_prefix}Add step {step_context} in scenario {scenario_context}")
                    scenario_context.add_step(step_context)
                
                # Process before step in reports
                try:
                    self.report_manager.before_step(step_context, step, step_level)
                except:
                    logger.exception(f"{log_prefix}Error while updating report before step for unstarted step")
            
            # Report
            try:
                step_level = scenario_context.scope_manager.scope_level("steps")
                self.report_manager.after_step(step_context, step, step_level)
            except:
                logger.exception(f"{log_prefix}Error while updating report after step")
    
    def has_step_context(self):
        return self.has_feature_context() and self.get_feature_context().has_scenario and self.get_scenario_context().has_step()
    
    def get_step_context(self):
        return self.get_scenario_context().get_current_step()
        


SessionContext.TSessionContext = SessionContext





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.element_exception import ElementNotFoundException
from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class FeatureClassResourceByName(object):
    __registered_resources_by_class_and_name = {}
    __ordered_resource_names_by_class = {}
    
    @classmethod
    def _get_class_resource_description(cls, plural=False):
        if plural:
            return 'class resources'
        else:
            return 'class resource'
    
    @classmethod
    def _clear_class_registered_resources(cls):
        cls_name = Typing.get_class_fullname(cls)
        if cls_name in cls.__registered_resources_by_class_and_name:
            cls.__registered_resources_by_class_and_name[cls_name].clear()
            cls.__ordered_resource_names_by_class[cls_name].clear()
    
    @classmethod
    def _get_class_registered_resources(cls):
        cls_name = Typing.get_class_fullname(cls)
        if cls_name in cls.__registered_resources_by_class_and_name:
            return dict(cls.__registered_resources_by_class_and_name[cls_name])
        else:
            return {}
    
    @classmethod
    def register_resource_in_class(cls, name, resource, index=-1):
        cls_name = Typing.get_class_fullname(cls)
        if cls_name not in cls.__registered_resources_by_class_and_name:
            cls.__registered_resources_by_class_and_name[cls_name] = {}
            cls.__ordered_resource_names_by_class[cls_name] = []
            
        cls.__registered_resources_by_class_and_name[cls_name][name] = resource
        
        while name in cls.__ordered_resource_names_by_class[cls_name]:
            cls.__ordered_resource_names_by_class[cls_name].remove(name)
        if index < 0:
            cls.__ordered_resource_names_by_class[cls_name].append(name)
        else:
            cls.__ordered_resource_names_by_class[cls_name].insert(index, name)
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{Typing.get_class_fullname(cls)}] Registered class resources: {cls.__ordered_resource_names_by_class[cls_name]}")
    
    @classmethod
    def import_registered_resources_in_class(cls, class_=None, obj=None, index=-1):
        if class_ is not None and issubclass(class_, FeatureClassResourceByName):
            resources_to_import_by_name = class_._get_class_registered_resources()
            ordered_resource_names_to_import = class_.get_class_resource_names()
        elif obj is not None and isinstance(obj, FeatureClassResourceByName):
            resources_to_import_by_name = obj._get_class_registered_resources()
            ordered_resource_names_to_import = obj.get_class_resource_names()
        elif obj is not None and isinstance(obj, FeatureObjectResourceByName):
            resources_to_import_by_name = obj._registered_resources
            ordered_resource_names_to_import = obj.get_resource_names()
        else:
            raise TechnicalException(f"Unmanaged import registered {cls._get_class_resource_description(True)} from (class_={class_} ; type(obj)={Typing.get_object_class_fullname(obj)})")
        
        for name in ordered_resource_names_to_import:
            cls.register_resource_in_class(name, resources_to_import_by_name[name], index=index)
        
    @classmethod
    def get_class_resource(cls, name, raise_if_not_found=True):
        registered_resources = cls._get_class_registered_resources()
        if name in registered_resources:
            return registered_resources[name]
        
        if raise_if_not_found:
            raise ElementNotFoundException(f"Not found {cls._get_class_resource_description()} of name '{name}' in registered {cls._get_class_resource_description(True)} {list(registered_resources.keys())}")
        else:
            return None
        
    @classmethod
    def get_class_resource_names(cls):
        cls_name = Typing.get_class_fullname(cls)
        if cls_name in cls.__ordered_resource_names_by_class:
            return list(cls.__ordered_resource_names_by_class[cls_name])
        else:
            return []


class FeatureObjectResourceByName(object):
    def __init__(self):
        self.__registered_resources_by_name = {}
        self.__ordered_resource_names = []
    
    @property
    def _registered_resources(self):
        return dict(self.__registered_resources_by_name)
    
    def _get_resource_description(self, plural=False):
        if plural:
            return 'resources'
        else:
            return 'resource'
    
    def _clear_registered_resources(self):
        self.__registered_resources_by_name.clear()
        self.__ordered_resource_names.clear()
    
    def register_resource(self, name, resource, index=-1):
        self.__registered_resources_by_name[name] = resource
        
        while name in self.__ordered_resource_names:
            self.__ordered_resource_names.remove(name)
        if index < 0:
            self.__ordered_resource_names.append(name)
        else:
            self.__ordered_resource_names.insert(index, name)
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{Typing.get_object_class_fullname(self)}] Registered resources: {self.__ordered_resource_names}")
    
    def import_registered_resources(self, class_=None, obj=None, index=-1):
        if class_ is not None and issubclass(class_, FeatureClassResourceByName):
            resources_to_import_by_name = class_._get_class_registered_resources()
            ordered_resource_names_to_import = class_.get_class_resource_names()
        elif obj is not None and isinstance(obj, FeatureClassResourceByName):
            resources_to_import_by_name = obj._get_class_registered_resources()
            ordered_resource_names_to_import = obj.get_class_resource_names()
        elif obj is not None and isinstance(obj, FeatureObjectResourceByName):
            resources_to_import_by_name = obj._registered_resources
            ordered_resource_names_to_import = obj.get_resource_names()
        else:
            raise TechnicalException(f"Unmanaged import registered {self._get_resource_description(True)} from (class_={class_} ; type(obj)={Typing.get_object_class_fullname(obj)})")
        
        for name in ordered_resource_names_to_import:
            self.register_resource(name, resources_to_import_by_name[name], index=index)
    
    def get_resource(self, name, raise_if_not_found=True):
        if name in self.__registered_resources_by_name:
            return self.__registered_resources_by_name[name]
        
        if raise_if_not_found:
            raise ElementNotFoundException(f"Not found {self._get_resource_description()} of name '{name}' in registered {self._get_resource_description(True)} {list(self.__registered_resources_by_name.keys())}")
        else:
            return None
    
    def get_resource_names(self):
        return list(self.__ordered_resource_names)


class FeatureObjectOverClassResourceByName(FeatureObjectResourceByName):
    def __init__(self):
        super().__init__()

    def get_resource(self, name, raise_if_not_found=True):
        res = super().get_resource(name, raise_if_not_found=False)
        if not res:
            res = self.get_class_resource(name, raise_if_not_found=False)
        
        if not res and raise_if_not_found:
            raise ElementNotFoundException(f"Not found {self._get_resource_description()} of name '{name}' in registered {self._get_resource_description(True)} {list(self._registered_resources.keys())} and registered {self._get_class_resource_description(True)} {list(self._get_class_registered_resources().keys())}")
        return res
    
    def get_resource_names(self):
        res = super().get_resource_names()
        class_resource_names = self.get_class_resource_names()
        for name in class_resource_names:
            res.append(name)
        return res


class FeatureObjectAndClassResourceByName(FeatureObjectOverClassResourceByName, FeatureClassResourceByName):
    def __init__(self):
        FeatureObjectOverClassResourceByName.__init__(self)






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from typing import NamedTuple
import threading
from holado.common.handlers.object import DeleteableObject, Object
from holado.common.handlers.enums import ObjectStates
from holado_multitask.multithreading.functionthreaded import FunctionThreaded
from holado_core.common.handlers.redo import Redo
from holado_system.system.filesystem.file import File
from holado_core.common.exceptions.verify_exception import VerifyException
import abc
from holado_multitask.multithreading.loopfunctionthreaded import LoopFunctionThreaded
from holado_multitask.multitasking.multitask_manager import MultitaskManager


logger = logging.getLogger(__name__)

try:
    import pika  # @UnresolvedImport @UnusedImport
    from pika.exceptions import ConnectionWrongStateError, StreamLostError
    with_pika = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"RMQClient is not available. Initialization failed on error: {exc}")
    with_pika = False



class RMQClient(DeleteableObject):
    __metaclass__ = abc.ABCMeta
    
    @classmethod
    def is_available(cls):
        return with_pika
    
    def __init__(self, name):
        super().__init__(name)
        
        self.__connection_kwargs = None
        self.__publisher_and_kwargs = []
        self.__consumer_and_kwargs = []
        self.__buffer_consumer_and_kwargs = []
        
        self.__connection = None
        self.__connection_thread_id = None
        
        self.__is_consuming = False
        self.__is_stopping_consuming = False
        self.__rapid_close = False
        
        self.__process_data_events_thread = None
        self.__is_processing_data_events = False
        self.__is_stopping_processing_data_events = False
    
    def _delete_object(self):
        try:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Deleting RabbitMQ client...")
            self.close()
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Finished to delete RabbitMQ client")
        except StreamLostError as exc:
            if self.object_state == ObjectStates.Deleting:
                pass
            else:
                raise exc
        
    @property
    def connection(self):
        return self.__connection
    
    @property
    def _connection_kwargs(self):
        return self.__connection_kwargs
    
    @property
    def is_in_connection_thread(self):
        return MultitaskManager.get_thread_id() == self.__connection_thread_id
    
    @property
    def is_consuming(self):
        return self.__is_consuming
    
    @property
    def is_processing_data_events(self):
        return self.__is_processing_data_events
    
    @property
    def rapid_close(self):
        return self.__rapid_close
    
    @rapid_close.setter
    def rapid_close(self, rapid_close):
        self.__rapid_close = rapid_close
    
    @property
    def has_consumer(self):
        return len(self.__consumer_and_kwargs) > 0
    
    @property
    def _consumer_and_kwargs(self):
        return self.__consumer_and_kwargs
    
    @property
    def has_buffer_consumer(self):
        return len(self.__buffer_consumer_and_kwargs) > 0
    
    @property
    def _buffer_consumer_and_kwargs(self):
        return self.__buffer_consumer_and_kwargs
    
    @property
    def has_publisher(self):
        return len(self.__publisher_and_kwargs) > 0
    
    @property
    def _publisher_and_kwargs(self):
        return self.__publisher_and_kwargs
    
    def _set_connection(self, connection):
        self.__connection = connection
        self.__connection_thread_id = MultitaskManager.get_thread_id()
        
    def _build_and_save_connection_parameters_kwargs(self, **connection_parameters_kwargs):
        if Tools.has_sub_kwargs(connection_parameters_kwargs, "authentication."):
            authentication = Tools.pop_sub_kwargs(connection_parameters_kwargs, "authentication.")
            if 'user' in authentication:
                if type(authentication['user']) is tuple:
                    connection_parameters_kwargs['credentials'] = pika.PlainCredentials(authentication['user'][0], authentication['user'][1])
                else:
                    raise FunctionalException(f"[{self.name}]When authenticating by user, the value has to be in format: ('{{USER}}', '{{PASSWORD}}')  (obtained: {authentication['user']})")
            else:
                raise TechnicalException(f"[{self.name}]Unmanaged authentication type '{authentication.keys()}' (possible authentication types: 'user'")

        self.__connection_kwargs = connection_parameters_kwargs
        
    def _new_connection_parameters(self):
        return pika.ConnectionParameters(**self._connection_kwargs)
    
    @abc.abstractmethod
    def _new_connection(self):
        raise NotImplementedError()
    
    def connect(self, **connection_parameters_kwargs):
        """Connect with client appropriate connection"""
        self._build_and_save_connection_parameters_kwargs(**connection_parameters_kwargs)
        connection = self._new_connection()
        self._set_connection(connection)
    
    def close(self):
        # Stop consuming
        if self.is_consuming:
            self.stop_consuming()
            
        # Stop processing data events
        if self.is_processing_data_events:
            self.stop_process_data_events()
            
        # Close connection
        if self.__connection and self.__connection.is_open:
            self._close_connection()
        
    def _close_connection(self, connection=None):
        if connection is None:
            connection = self.connection
            
        if connection is None or not connection.is_open:
            raise TechnicalException(f"[{self.name}] Connection is not opened")
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Closing connection ({id(connection)})...")
        
        try:
            self.__connection.close()
        except (StreamLostError, ConnectionWrongStateError):
            pass
        except Exception as exc:  # @UnusedVariable
            #TODO: When this warning is logged during self.__del__, the log is cleared before, thus it is commented
            # logger.warn(f"Error catched while closing RabbitMQ client connection:\n{Tools.represent_exception(exc)}")
            # pass
            raise exc
        
    def __new_object_kwargs(self, obj, **kwargs):
        res = NamedTuple('ObjectKwargs', object=object, kwargs=dict)
        res.object = obj
        res.kwargs = kwargs
        return res

    def new_publisher(self, queue, queue_args=None, exchange="", exchange_args=None, routing_key=None, nb_runners=None):
        return self._new_publisher_with_connection(self.connection, queue, queue_args, exchange, exchange_args, routing_key, nb_runners)

    def _new_publisher_with_connection(self, connection, queue, queue_args=None, exchange="", exchange_args=None, routing_key=None, nb_runners=None):
        pub_channel = connection.channel()
        
        queue_name = self._prepare_queue(pub_channel, queue, queue_args, exchange, exchange_args, bind_args=None)
        if self.connection.publisher_confirms:
            pub_channel.set_confirm_delivery()
        else:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Connection doesn't support publisher confirms")

        if routing_key:
            res = self._new_publisher(connection, pub_channel, exchange, routing_key, nb_runners)
        else:
            # TODO EKL: verify this case is functional
            res = self._new_publisher(connection, pub_channel, exchange, queue_name, nb_runners)
            
        self.__publisher_and_kwargs.append( self.__new_object_kwargs(res, queue=queue, queue_args=queue_args, exchange=exchange, exchange_args=exchange_args, routing_key=routing_key, nb_runners=nb_runners) )
        return res
    
    def _new_publisher(self, connection, channel, exchange, routing_key, nb_runners):
        raise NotImplementedError
        
    def new_consumer(self, queue, message_callback, queue_args=None, exchange="", exchange_args=None, bind_args=None):
        return self._new_consumer_with_connection(self.connection, queue, message_callback, queue_args, exchange, exchange_args, bind_args)
        
    def _new_consumer_with_connection(self, connection, queue, message_callback, queue_args=None, exchange="", exchange_args=None, bind_args=None):
        con_channel = connection.channel()
        
        if self.is_consuming:
            raise FunctionalException(f"[{self.name}] Not allowed to create a new consumer while consuming is started")
        queue_name = self._prepare_queue(con_channel, queue, queue_args, exchange, exchange_args, bind_args)
        res = RMQConsumer(self, connection, con_channel, queue_name, message_callback)
        self.__consumer_and_kwargs.append( self.__new_object_kwargs(res, queue=queue, message_callback=message_callback, queue_args=queue_args, exchange=exchange, exchange_args=exchange_args, bind_args=bind_args) )
        return res
        
    def new_buffer_consumer(self, queue, queue_args=None, exchange="", exchange_args=None, bind_args=None):
        return self._new_buffer_consumer_with_connection(self.connection, queue, queue_args, exchange, exchange_args, bind_args)
        
    def _new_buffer_consumer_with_connection(self, connection, queue, queue_args=None, exchange="", exchange_args=None, bind_args=None):
        con_channel = connection.channel()
        
        if self.is_consuming:
            raise FunctionalException(f"[{self.name}] Not allowed to create a new consumer while consuming is started")
        queue_name = self._prepare_queue(con_channel, queue, queue_args, exchange, exchange_args, bind_args)
        res = RMQBufferConsumer(self, connection, con_channel, queue_name)
        self.__buffer_consumer_and_kwargs.append( self.__new_object_kwargs(res, queue=queue, queue_args=queue_args, exchange=exchange, exchange_args=exchange_args, bind_args=bind_args) )
        return res
        
    def start_consuming(self):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Start consuming...")
        self.__is_consuming = True
        
        for obj_kwargs in self.__consumer_and_kwargs:
            obj_kwargs.object.start_consuming_in_thread()
            
        for obj_kwargs in self.__buffer_consumer_and_kwargs:
            obj_kwargs.object.start_consuming_in_thread()
        
    def stop_consuming(self):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Beginning stop consuming...")
        self.__is_stopping_consuming = True
        consumers_exceptions = []
        try:
            for obj_kwargs in self.__consumer_and_kwargs:
                try:
                    obj_kwargs.object.stop_consuming()
                except Exception as exc:
                    consumers_exceptions.append( (obj_kwargs.object, exc) )
            for obj_kwargs in self.__buffer_consumer_and_kwargs:
                try:
                    obj_kwargs.object.stop_consuming()
                except Exception as exc:
                    consumers_exceptions.append( (obj_kwargs.object, exc) )
        finally:
            self.__is_stopping_consuming = False
            self.__is_consuming = False
            
            if consumers_exceptions:
                logger.error(f"[{self.name}] Errors while stopping consuming on consumers: { {o.name:e for o,e in consumers_exceptions} }")
            elif Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Finished stop consuming")

    def start_process_data_events(self, raise_if_already_started=True):
        if self.__process_data_events_thread is not None and self.__process_data_events_thread.is_alive():
            if raise_if_already_started:
                raise TechnicalException(f"[{self.name}] Thread processing data events is already running")
            else:
                return
            
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Start processing data events...")
        self.__is_processing_data_events = True
        
        self.__process_data_events_thread = LoopFunctionThreaded(self.connection.process_data_events, kwargs={'time_limit':1}, register_thread=True, delay_before_run_sec=None, delay_between_run_sec=0.01)
        self.__process_data_events_thread.start()

    def stop_process_data_events(self):
        if not self.__is_processing_data_events:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Client is not processing data events")
            return
        if self.__is_stopping_processing_data_events:
            raise TechnicalException(f"[{self.name}] Data events processing is already under stop")
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Beginning stop data events processing...")
        self.__is_stopping_processing_data_events = True
        try:
            self.__process_data_events_thread.interrupt()
            self.__process_data_events_thread.join()
        finally:
            self.__is_stopping_processing_data_events = False
            self.__is_processing_data_events = False
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Finished stop data events processing")

    def _prepare_queue(self, channel, queue, queue_args=None, exchange="", exchange_args=None, bind_args=None):
        if exchange is not None and exchange != "":
            self._exchange_declare(channel, exchange, exchange_args)
        
        if queue is not None and queue != "":
            self._queue_declare(channel, queue, queue_args)
            queue_name = queue
        else:
            q_args = dict(queue_args) if queue_args else {}
            q_args['exclusive'] = True
            result = self._queue_declare(channel, queue, q_args)
            queue_name = result.method.queue
            
        if exchange is not None and exchange != "":
            self._queue_bind(channel, queue_name, exchange, bind_args)
            
        return queue_name
    
    def delete_queue(self, queue):
        self.connection.channel().queue_delete(queue=queue)
    
    def purge_queue(self, queue):
        self.connection.channel().queue_purge(queue)
        
    def _exchange_declare(self, channel, exchange, exchange_args=None):
        kwargs = {}
        arguments = None
        if exchange_args:
            arguments = dict(exchange_args)
            for name in ['exchange_type', 'passive', 'durable', 'auto_delete', 'internal']:
                if name in arguments:
                    kwargs[name] = arguments.pop(name)
        return channel.exchange_declare(exchange, arguments=arguments, **kwargs)
        
    def _queue_declare(self, channel, queue, queue_args=None):
        kwargs = {}
        arguments = None
        if queue_args:
            arguments = dict(queue_args)
            for name in ['passive', 'durable', 'exclusive', 'auto_delete']:
                if name in arguments:
                    kwargs[name] = arguments.pop(name)
        return channel.queue_declare(queue, arguments=arguments, **kwargs)
        
    def _queue_bind(self, channel, queue, exchange, bind_args=None):
        kwargs = {}
        arguments = None
        if bind_args:
            arguments = dict(bind_args)
            for name in ['routing_key']:
                if name in arguments:
                    kwargs[name] = arguments.pop(name)
        return channel.queue_bind(queue, exchange, arguments=arguments, **kwargs)
    
    def get_queue_message_count(self, queue):
        channel = self.connection.channel()
        status = self._queue_declare(channel, queue, {'passive':True})
        return status.method.message_count
    
    def is_queue_empty(self, queue, raise_exception=False):
        nb = self.get_queue_message_count(queue)
        res = (nb == 0)
        
        if not res and raise_exception:
            raise VerifyException(f"[{self.name}] Queue '{queue}' is not empty, it contains {nb} messages.")
        return res
    
    def flush(self):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Flushing...")
        for obj_kwargs in self._publisher_and_kwargs:
            obj_kwargs.object.flush()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Flushed")
    
    
    
class RMQPublisher(Object):
    def __init__(self, client, connection, channel, exchange, routing_key):
        super().__init__(f"RMQPublisher({routing_key})")
        self.__client = client
        self.__connection = connection
        self.__channel = channel
        self.__exchange = exchange
        self.__routing_key = routing_key
        
    @property
    def client(self):
        return self.__client
    
    @property
    def connection(self):
        return self.__connection
    
    @property
    def channel(self):
        return self.__channel
    
    def publish(self, body, **kwargs):
        kwargs['properties'] = pika.BasicProperties(delivery_mode=pika.DeliveryMode.Transient)
        # kwargs['properties'] = pika.BasicProperties(delivery_mode=pika.DeliveryMode.Persistent)
        # if self.__channel.has_confirm_delivery:
        #     kwargs['mandatory'] = True
        self.__channel.basic_publish(exchange=self.__exchange, routing_key=self.__routing_key, body=body, **kwargs)
        
        
class RMQConsumer(DeleteableObject):
    def __init__(self, client, connection, channel, queue, message_callback):
        super().__init__(f"RMQConsumer({queue})")
        
        self.__client = client
        self.__connection = connection
        self.__channel = channel
        self.__queue = queue
        self.__message_callback = message_callback
        
        # self.__consuming_thread_id = None
        self.__is_consuming = False
        self.__is_stopping_consuming = False
        
        self.__consumer_tag = self.__channel.basic_consume(queue=self.__queue, on_message_callback=self.__message_callback, auto_ack=True)
    
    @property
    def connection(self):
        return self.__connection
    
    @property
    def channel(self):
        return self.__channel
    
    @property
    def queue(self):
        return self.__queue
    
    @property
    def consumer_tag(self):
        return self.__consumer_tag
    
    # @property
    # def consuming_thread_id(self):
    #     return self.__consuming_thread_id
    #
    # @consuming_thread_id.setter
    # def consuming_thread_id(self, thread_id):
    #     self.__consuming_thread_id = thread_id
    
    @property
    def is_consuming(self):
        return self.__is_consuming
    
    def _delete_object(self):
        try:
            # Cancel consumer
            # self.__channel.basic_cancel(self.consumer_tag)
            
            # Stop consuming
            if self.is_consuming:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[{self.name}] Deleting RabbitMQClient: Stopping consuming...")
                self.stop_consuming()
            
            # Close channel
            if self.__channel.is_open:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[{self.name}] Deleting RabbitMQClient: Closing channel...")
                self.__channel.close()
        except StreamLostError as exc:
            if self.object_state == ObjectStates.Deleting:
                pass
            else:
                raise exc
    
    def start_consuming_in_thread(self):
        func = FunctionThreaded(self.start_consuming, register_thread=True, delay_before_run_sec=None)
        func.start()
        # self.consuming_thread_id = MultitaskManager.get_thread_id(thread=func)
        
    def start_consuming(self):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Beginning start consuming...")
        
        # Workaround: 
        #    Sometimes "When start consuming" is called rather than "Then client is consuming" 
        #    cf implementation of step @Given(r"start consuming in a thread \(RMQ client: (?P<var_client>\w+)\)")
        #    Workaround: register the thread_id allowed to start consuming.
        #    Note: A possible reason is that execute_steps is not thread safe. If it's true, the new implementation
        #          of the step with the wait of 0.01 seconds has resolved the problem, and thus this workaround is not needed anymore.
        # if self.consuming_thread_id is not None:
        #     thread_id = MultitaskManager.get_thread_id()
        #     if thread_id != self.consuming_thread_id:
        #         logger.error(f"Only thread {self.consuming_thread_id} is allowed to start consuming. Tried to start consuming in thread {thread_id} with traceback:\n{traceback.represent_stack(indent=4)}")
        #         return
        
        self.__is_consuming = True
        try:
            self.channel.start_consuming()
        except Exception as exc:
            if self.__is_stopping_consuming:
                if isinstance(exc, pika.exceptions.StreamLostError) \
                        or isinstance(exc, AttributeError) and "NoneType' object has no attribute 'clear'" in str(exc) \
                        or isinstance(exc, AssertionError):
                    logger.info(f"[{self.name}] Caught exception in consuming thread while stopping consuming: {exc}")
                else:
                    logger.warn(f"[{self.name}] Caught unexpected exception in consuming thread while stopping consuming: {Tools.represent_exception(exc)}")
            else:
                raise TechnicalException(f"[{self.name}] Failed to start consuming: {str(exc)}") from exc
        finally:
            # In all cases, self.__is_consuming must began False, otherwise stopping processing is broken in case of "raise exc"
            self.__is_consuming = False
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Finished start consuming")
    
    def stop_consuming(self):
        if not self.__is_consuming:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Consumer is not consuming")
            return
        if self.__is_stopping_consuming:
            raise TechnicalException(f"[{self.name}] Consuming is already under stop")
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Beginning stop consuming...")
        self.__is_stopping_consuming = True
        try:
            self.__stop_consuming_by_channel_stop_consuming()
            # self.__stop_consuming_by_deleting_channel_consumer_infos()
            
            # gc.collect()
            
            # Wait end of consuming
            if self.__is_consuming:
                logger.info(f"[{self.name}] Waiting consuming thread is stopped...")
                
                class StopRedo(Redo):
                    def __init__(self2):  # @NoSelf
                        super().__init__(f"[{self.name}] Stop consuming on consumer {self.consumer_tag}")
                        
                    def _process(self2):  # @NoSelf
                        return self.is_consuming
                redo = StopRedo()
                redo.redo_while(True)
                redo.with_timeout(30*60)    # Timeout of 30 min
                redo.polling_every(0.01)     # Wait 0.1 s beetween each try
                redo.with_process_in_thread(False)      # Run process in thread is not needed
                redo.execute()
                
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[{self.name}] Finished waiting consuming thread is stopped")
        finally:
            self.__is_stopping_consuming = False
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Finished stop consuming")
            
    def __stop_consuming_by_channel_stop_consuming(self):
        try:
            self.channel.stop_consuming()
        except Exception as exc:
            if isinstance(exc, pika.exceptions.StreamLostError) \
                    or isinstance(exc, AttributeError) and "NoneType' object has no attribute 'clear'" in str(exc) \
                    or isinstance(exc, AssertionError):
                logger.info(f"[{self.name}] Caught exception while executing consuming stopping method: {exc}")
            else:
                logger.warn(f"[{self.name}] Caught exception while executing consuming stopping method: {Tools.represent_exception(exc)}")

    def __stop_consuming_by_deleting_channel_consumer_infos(self):
        del self.channel._consumer_infos[self.consumer_tag]
        # Schedule termination of connection.process_data_events using a
        # negative channel number
#                self.connection._request_channel_dispatch(-self.channel.channel_number)
        self.client.connection._request_channel_dispatch(-self.channel.channel_number)
    
    
    
class RMQBufferConsumer(Object):
    def __init__(self, client, connection, channel, queue):
        super().__init__(f"RMQBufferConsumer({queue})")
        self.__consumer = RMQConsumer(client, connection, channel, queue, self.__message_callback)
        
        self.__messages = []
        self.__messages_lock = threading.Lock()
    
    @property
    def connection(self):
        return self.__consumer.connection
    
    @property
    def nb_messages(self):
        with self.__messages_lock:
            return len(self.__messages)
    
    @property
    def messages(self):
        with self.__messages_lock:
            return list(self.__messages)
    
    @property
    def consumer_tag(self):
        return self.__consumer.consumer_tag
    
    # @property
    # def consuming_thread_id(self):
    #     return self.__consumer.consuming_thread_id
    #
    # @consuming_thread_id.setter
    # def consuming_thread_id(self, thread_id):
    #     self.__consumer.consuming_thread_id = thread_id
        
    def start_consuming_in_thread(self):
        self.__consumer.start_consuming_in_thread()
        
    def start_consuming(self):
        self.__consumer.start_consuming()
    
    def stop_consuming(self):
        self.__consumer.stop_consuming()
        
    def __message_callback(self, channel, method, properties, body):
        with self.__messages_lock:
            self.__messages.append( (channel, method, properties, body) )
            # logger.debug(f"[Consumer '{self.__consumer.queue}'] New message (total: {len(self.__messages)}): {channel=} ; {method=} ; {properties=} ; {body=}")
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"[{self.name}] New message (total: {len(self.__messages)}): {body}")
            elif Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] New message (total: {len(self.__messages)})")
    
    def reset_messages(self):
        with self.__messages_lock:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Reset messages (delete {len(self.__messages)} messages)")
            self.__messages.clear()
        
    def pop_first_message(self):
        with self.__messages_lock:
            res = self.__messages.pop(0)
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Pop first message (remaining: {len(self.__messages)}): {res}")
            return res
        
    def pop_and_save_messages(self, file:File, max_messages=None):
        res = 0
        
        nb_messages = self.nb_messages
        while nb_messages > 0 and (max_messages is None or res < nb_messages):
            msg = self.pop_first_message()
            file.writelines([msg[3].hex()])
            res += 1
            
            nb_messages = self.nb_messages
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Saved a message (nb saved: {res} ; remaining: {nb_messages})")
            
        return res
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_report.report.builders.report_builder import ReportBuilder
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.technical_exception import TechnicalException
import json
from holado_system.system.filesystem.file import File
from holado_xml.xml.xml_file import XMLFile

logger = logging.getLogger(__name__)



class FailureReportBuilder(ReportBuilder):
    """Failure report builder
    Supported formats: 'txt', 'json', 'xml'
    """
    def __init__(self, filepath, file_format='xml', sort_by_nb_scenario=True, exclude_statuses=None, exclude_categories=None):
        self.__file_format = file_format.lower()
        self.__sort_by_nb_scenario = sort_by_nb_scenario
        self.__exclude_statuses = exclude_statuses
        self.__exclude_categories = exclude_categories
        
        if self.__file_format in ['txt', 'json']:
            self.__file = File(filepath, mode='wt')
        elif self.__file_format == 'xml':
            self.__file = XMLFile(filepath, mode='wt')
        else:
            raise TechnicalException(f"Unmanaged format '{self.__file_format}' (possible formats: 'txt', 'json', 'xml')")
        
        self.__failures = {}
        
    def build(self):
        '''
        The file is built after each scenario
        '''
        pass
        
    def after_scenario(self, scenario, scenario_report=None):
        from holado_report.report.report_manager import ReportManager
        category_validation, status_validation, step_failed, step_number, scenario_context, step_context = ReportManager.get_current_scenario_status_information(scenario)
        
        # Manage excluded scenarios
        if self.__exclude_statuses and status_validation in self.__exclude_statuses:
            return
        if category_validation is not None and self.__exclude_categories:
            ind = category_validation.find(' (')
            category = category_validation[:ind] if ind > 0 else category_validation
            if category in self.__exclude_categories:
                return
        
        self.__file_add_scenario(scenario, scenario_report, category_validation, status_validation, step_failed, step_number, scenario_context, step_context)
    
    def __file_add_scenario(self, scenario, scenario_report, category_validation, status_validation, step_failed, step_number, scenario_context, step_context):
        if status_validation != "Passed" and step_failed is not None:
            self.__add_failure(scenario, scenario_report, category_validation, status_validation, step_failed, step_number, scenario_context, step_context)
            self.__update_file()
            
    def __add_failure(self, scenario, scenario_report, category_validation, status_validation, step_failed, step_number, scenario_context, step_context):
        from holado_report.report.report_manager import ReportManager
        
        step_error_message = ReportManager.get_step_error_message(step_failed).strip()
        if step_error_message not in self.__failures:
            self.__failures[step_error_message] = []
        
        category_str = f" => {category_validation}" if category_validation else ""
        if self.__file_format == 'txt':
            msg_list = []
            msg_list.append(f"scenario in {ReportManager.format_scenario_short_description(scenario)} - {ReportManager.format_step_short_description(step_failed, step_number, has_failed=True)} - {status_validation}{category_str}")
            msg_list.append(f"    Feature/Scenario: {scenario.feature.name}  =>  {scenario.name}")
            msg_list.append(f"    Report: {scenario_report.report_path}")
            msg_list.append(f"    Tags: -t " + " -t ".join(scenario.feature.tags + scenario.tags))
            if step_context and step_context.start_datetime is not None:
                msg_list.append(f"    Scenario/Step periods: {ReportManager.format_context_period(scenario_context)} -> {ReportManager.format_context_period(step_context)}")
            else:
                msg_list.append(f"    Scenario period: {ReportManager.format_context_period(scenario_context)}")
            msg_scenario = "\n".join(msg_list)
            
            self.__failures[step_error_message].append(msg_scenario)
        elif self.__file_format in ['json', 'xml']:
            scenario_info = {
                'title': f"{ReportManager.format_scenario_short_description(scenario)} - {ReportManager.format_step_short_description(step_failed, step_number, has_failed=True)} - {status_validation}{category_str}",
                'scenario': f"{scenario.feature.name}  =>  {scenario.name}",
                'report': scenario_report.report_path,
                'tags': "-t " + " -t ".join(scenario.feature.tags + scenario.tags)
                }
            if step_context and step_context.start_datetime is not None:
                scenario_info['periods'] = f"{ReportManager.format_context_period(scenario_context)} -> {ReportManager.format_context_period(step_context)}"
            else:
                scenario_info['periods'] = f"{ReportManager.format_context_period(scenario_context)}"
                
            self.__failures[step_error_message].append(scenario_info)
        else:
            raise TechnicalException(f"Unmanaged format '{self.__file_format}' (possible formats: 'txt', 'json')")
    
    def __update_file(self):
        failures = dict(sorted(self.__failures.items(), key=lambda item:-len(item[1]))) if self.__sort_by_nb_scenario else self.__failures
        
        with self.__file as fout:
            if self.__file_format == 'txt':
                for failure, scenarios_messages in failures.items():
                    fout.write(failure + "\n")
                    fout.write("\n")
                    for msg in scenarios_messages:
                        fout.write(Tools.indent_string(4, msg) + "\n")
                        fout.write("\n")
            elif self.__file_format == 'json':
                json_str = json.dumps(failures, ensure_ascii=False, indent=4)
                fout.write(json_str)
            elif self.__file_format == 'xml':
                self.__file_write_failures_xml(fout, failures)
            else:
                raise TechnicalException(f"Unmanaged format '{self.__file_format}' (possible formats: 'txt', 'json')")

    def __file_write_failures_xml(self, fout, failures):
        output_dict = {
            'failures': {
                'failure': [self.__convert_failure_to_output_dict(error_message, scenarios) for error_message, scenarios in failures.items()]
                }
            }
        
        fout.write_dict(output_dict, pretty=True, indent=Tools.indent_string(4, ''))
    
    def __convert_failure_to_output_dict(self, error_message, scenarios):
        res = {}
        
        if "\n" in error_message:
            res['error_message'] = "\n" + Tools.indent_string(12, error_message) + Tools.indent_string(8, "\n")
        else:
            res['error_message'] = error_message
        
        res['scenarios'] = {'scenario': scenarios}
        
        return res
    






# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_python.standard_library.socket.blocking_socket import TCPBlockingSocketClient
from holado_python.standard_library.socket.echo_server import EchoTCPBlockingSocketServer
from holado_python.standard_library.socket.message_socket import MessageTCPNonBlockingSocketClient, MessageTCPBlockingSocketClient
from holado_core.common.handlers.wait import WaitEndChange, WaitChange
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_python.standard_library.socket.non_blocking_socket import TCPNonBlockingSocketClient

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()



################################################################
## Clients
################################################################


@Given(r"(?P<var_name>{Variable}) = new TCP socket client")
def step_impl(context, var_name):
    """Return a new TCP socket client.
    Note: Only IPv4 is managed for the moment
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)

    blocking = kwargs.pop('blocking') if 'blocking' in kwargs else True
    if blocking:
        res = TCPBlockingSocketClient(create_ipv4_socket_kwargs=kwargs)
    else:
        res = TCPNonBlockingSocketClient(create_ipv4_socket_kwargs=kwargs)
    
    __get_variable_manager().register_variable(var_name, res)


@Given(r"(?P<var_name>{Variable}) = new message TCP socket client")
def step_impl(context, var_name):
    """Return a new message TCP socket client.
    Parameter 'separator' specifies the separator at end of each message (default: b'\n')
    Note: Only IPv4 is managed for the moment
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
    
    separator = kwargs.pop('separator') if 'separator' in kwargs else '\n'
    blocking = kwargs.pop('blocking') if 'blocking' in kwargs else True
    if blocking:
        res = MessageTCPBlockingSocketClient(separator, create_ipv4_socket_kwargs=kwargs)
    else:
        res = MessageTCPNonBlockingSocketClient(separator, create_ipv4_socket_kwargs=kwargs)
    
    __get_variable_manager().register_variable(var_name, res)


@Step(r"start \(socket client: (?P<server_varname>{Variable})\)")
def step_impl(context, server_varname):
    """Start a socket client.
    It starts a thread listening incoming data.
    This step is needed for non blocking socket clients to manage incoming data.
    Note: current implementation uses 'start' method parameters default values.
    """
    server = StepTools.evaluate_scenario_parameter(server_varname)
    server.start()

@Step(r"stop \(socket client: (?P<server_varname>{Variable})\)")
def step_impl(context, server_varname):
    """Stop a non blocking socket client.
    """
    server = StepTools.evaluate_scenario_parameter(server_varname)
    server.stop()




################################################################
## Server
################################################################


@Given(r"(?P<var_name>{Variable}) = new echo TCP socket server")
def step_impl(context, var_name):
    """Return a new echo TCP socket server.
    When started, it waits connections, and for each connection it listen data and send them back
    Note: Only IPv4 is managed for the moment
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
    
    res = EchoTCPBlockingSocketServer(create_ipv4_socket_kwargs=kwargs)
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"start \(socket server: (?P<server_varname>{Variable})\)")
def step_impl(context, server_varname):
    """Start a socket server
    Note: current implementation uses 'start' method parameters default values.
    """
    server = StepTools.evaluate_scenario_parameter(server_varname)
    server.start()

@Step(r"stop \(socket server: (?P<server_varname>{Variable})\)")
def step_impl(context, server_varname):
    """Start a socket server
    Note: current implementation uses 'start' method parameters default values.
    """
    server = StepTools.evaluate_scenario_parameter(server_varname)
    server.stop()



################################################################
## Client and Server
################################################################


@Step(r"ensure SSL handshake is done \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, socket_varname):
    """Ensure SSL handshake is done.
    """
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    sock.ensure_ssl_handshake_is_done()

@Step(r"write (?P<data>{Bytes}) \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, data, socket_varname):
    """Send data
    This step do as many socket send as needed to send the whole data.
    """
    data = StepTools.evaluate_scenario_parameter(data)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    sock.write(data)

@Step(r"(?P<var_name>{Variable}) = read \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, var_name, socket_varname):
    """Read data
    This step do as many socket recv as needed to receive the whole data.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    res = sock.read()
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = received data \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, var_name, socket_varname):
    """Get currently received data
    This step works only with clients that are started, so that listening thread is started.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    res = sock.received_data
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = size of received data \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, var_name, socket_varname):
    """Get received messages without removing from socket
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    res = sock.received_data_size
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"socket (?P<socket_varname>{Variable}) doesn't receive any data(?: \((?:timeout: (?P<timeout_sec>{Float}) s)?(?: ; )?(?:polling: (?P<polling_sec>{Float}) s)?\))?")
def step_impl(context, socket_varname, timeout_sec, polling_sec):
    """Wait and verify socket doesn't receive any data
    """
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    timeout_sec = StepTools.evaluate_scenario_parameter(timeout_sec)
    polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
    
    if timeout_sec is None:
        timeout_sec = 10
    
    wait = WaitChange(f"Verify no data is received (socket: {sock})",
                      lambda: sock.received_data_size,
                      timeout_seconds=timeout_sec, polling_seconds=polling_sec)
    wait.without_raise_on_timeout()
    data_size = wait.execute()
    if data_size > 0:
        raise FunctionalException(f"[{sock.name}] Socket has received data of size {data_size}")

@Step(r"await socket (?P<socket_varname>{Variable}) receives a data(?: \((?:timeout: (?P<timeout_sec>{Float}) s)?(?: ; )?(?:polling: (?P<polling_sec>{Float}) s)?\))?")
def step_impl(context, socket_varname, timeout_sec, polling_sec):
    """Wait socket receives at least one data
    """
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    timeout_sec = StepTools.evaluate_scenario_parameter(timeout_sec)
    polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
    
    if timeout_sec is None:
        timeout_sec = 10
    
    wait = WaitChange(f"Wait data reception (socket: {sock})",
                      lambda: sock.received_data_size,
                      timeout_seconds=timeout_sec, polling_seconds=polling_sec)
    wait.without_raise_on_timeout()
    data_size = wait.execute()
    if data_size == 0:
        raise FunctionalException(f"[{sock.name}] No data was received (timeout: {wait.timeout} seconds)")

@Step(r"await socket (?P<socket_varname>{Variable}) receives data(?: \((?:first timeout: (?P<first_timeout_sec>{Float}) s)?(?: ; )?(?:window: (?P<window_sec>{Float}) s)?(?: ; )?(?:polling: (?P<polling_sec>{Float}) s)?\))?")
def step_impl(context, socket_varname, first_timeout_sec, window_sec, polling_sec):
    """Wait socket stop to receive data
    """
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    first_timeout_sec = StepTools.evaluate_scenario_parameter(first_timeout_sec)
    window_sec = StepTools.evaluate_scenario_parameter(window_sec)
    polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
    
    wait = WaitEndChange(f"Wait end of data reception (socket: {sock})",
                         lambda: sock.received_data_size,
                         first_timeout_seconds=first_timeout_sec, window_seconds=window_sec, polling_seconds=polling_sec)
    wait.without_raise_on_timeout()
    data_size = wait.execute()
    if data_size == 0:
        raise FunctionalException(f"[{sock.name}] No data was received (timeout: {wait.timeout} seconds)")

@Given(r"reset stored received data in socket (?P<socket_varname>{Variable})")
def step_impl(context, socket_varname):
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    sock.reset_received_data()



# For client and servers managing messages

@Step(r"write message (?P<data>{Bytes}) \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, data, socket_varname):
    """Send message
    """
    data = StepTools.evaluate_scenario_parameter(data)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    sock.write_message(data)

@Step(r"(?P<var_name>{Variable}) = read message \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, var_name, socket_varname):
    """Read a received message.
    If no message is in socket, None is returned.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    res = sock.read_message()
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = read messages \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, var_name, socket_varname):
    """Read all received messages.
    If no message is in socket, an empty list is returned.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    res = sock.read_messages()
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = received messages \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, var_name, socket_varname):
    """Get received messages without removing from socket
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    res = sock.messages
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = number of received messages \(socket: (?P<socket_varname>{Variable})\)")
def step_impl(context, var_name, socket_varname):
    """Get received messages without removing from socket
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    
    res = sock.nb_messages
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"socket (?P<socket_varname>{Variable}) doesn't receive any message(?: \((?:timeout: (?P<timeout_sec>{Float}) s)?(?: ; )?(?:polling: (?P<polling_sec>{Float}) s)?\))?")
def step_impl(context, socket_varname, timeout_sec, polling_sec):
    """Wait and verify socket doesn't receive any message
    """
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    timeout_sec = StepTools.evaluate_scenario_parameter(timeout_sec)
    polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
    
    if timeout_sec is None:
        timeout_sec = 10
    
    wait = WaitChange(f"Wait end of message reception (socket: {sock})",
                      lambda: sock.nb_messages,
                      timeout_seconds=timeout_sec, polling_seconds=polling_sec)
    wait.without_raise_on_timeout()
    nb_messages = wait.execute()
    if nb_messages > 0:
        raise FunctionalException(f"[{sock.name}] Socket has received {nb_messages} messages (timeout: {wait.timeout} seconds)")

@Step(r"await socket (?P<socket_varname>{Variable}) receives (?:(?P<no_str>no) )?messages(?: \((?:first timeout: (?P<first_timeout_sec>{Float}) s)?(?: ; )?(?:window: (?P<window_sec>{Float}) s)?(?: ; )?(?:polling: (?P<polling_sec>{Float}) s)?\))?")
def step_impl(context, socket_varname, no_str, first_timeout_sec, window_sec, polling_sec):
    """Wait until socket receives messages or no messages.
    Wait during max first_timeout_sec duration for a first message, then, and after each new message, wait during max window_sec duration for a new message.
    If no message is received during first_timeout_sec duration, an error is raised only if messages are waited (ie 'no ' is not added when calling step).
    """
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    no_messages = no_str is not None
    first_timeout_sec = StepTools.evaluate_scenario_parameter(first_timeout_sec)
    window_sec = StepTools.evaluate_scenario_parameter(window_sec)
    polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
    
    wait = WaitEndChange(f"Wait until socket receives {no_str if no_str is not None else ''}messages (socket: {sock})",
                         lambda: sock.nb_messages,
                         first_timeout_seconds=first_timeout_sec, window_seconds=window_sec, polling_seconds=polling_sec)
    wait.without_raise_on_timeout()
    nb_messages = wait.execute()
    if not no_messages and nb_messages == 0:
        raise FunctionalException(f"[{sock.name}] No message was received (timeout: {wait.timeout} seconds)")

@Step(r"await socket (?P<socket_varname>{Variable}) receives a message(?: \((?:timeout: (?P<timeout_sec>{Float}) s)?(?: ; )?(?:polling: (?P<polling_sec>{Float}) s)?\))?")
def step_impl(context, socket_varname, timeout_sec, polling_sec):
    """Get received messages without removing from socket
    """
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    timeout_sec = StepTools.evaluate_scenario_parameter(timeout_sec)
    polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
    
    if timeout_sec is None:
        timeout_sec = 10
    
    wait = WaitChange(f"Wait a message reception (socket: {sock})",
                      lambda: sock.nb_messages,
                      timeout_seconds=timeout_sec, polling_seconds=polling_sec)
    wait.without_raise_on_timeout()
    nb_messages = wait.execute()
    if nb_messages == 0:
        raise FunctionalException(f"[{sock.name}] No message was received (timeout: {wait.timeout} seconds)")

@Given(r"reset stored received messages in socket (?P<socket_varname>{Variable})")
def step_impl(context, socket_varname):
    sock = StepTools.evaluate_scenario_parameter(socket_varname)
    sock.reset_received_messages()
    






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from holado_value.common.tables.value_table import ValueTable
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_core.common.tables.table_row import TableRow
from holado_core.common.tables.converters.table_converter import TableConverter
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tables.table import Table
import copy
import logging
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_value.common.tools.value_types import ValueTypes
from holado_core.common.tools.tools import Tools
from holado_json.ipc.json_converter import JsonConverter
from holado_python.standard_library.typing import Typing
from holado_data.data.generator.base import BaseGenerator

logger = logging.getLogger(__name__)


class ValueTableConverter(TableConverter):
    
    @classmethod
    def convert_table_2_value_table(cls, table, do_eval_once=True):
        if isinstance(table, TableWithHeader):
            res = ValueTableWithHeader()
            res.header = copy.copy(table.header)
        else:
            res = ValueTable()
        for row in table:
            cells_content = [f"'{cc}'" if isinstance(cc, str) else f"{cc}" for cc in row.cells_content]
            res.add_row(cells_content=cells_content, do_eval_once=do_eval_once)
        return res
    
    @classmethod
    def convert_value_table_2_table(cls, table):
        ValueTableManager.verify_is_value_table(table)
        
        if isinstance(table, ValueTableWithHeader):
            res = TableWithHeader()
            res.header = copy.copy(table.header)
        else:
            res = Table()
        for row in table:
            cells_content = row.cells_value
            res.add_row(cells_content=cells_content)
        return res
    
    @classmethod
    def convert_table_with_header_to_dict_list(cls, table, as_generator=False):
        if not ValueTableManager.is_value_table(table):
            return TableConverter.convert_table_with_header_to_dict_list(table, as_generator=as_generator)
        
        if as_generator:
            class TableWithHeader2DictGenerator(BaseGenerator):
                def __init__(self, table):
                    super().__init__(name="table with header to dict generator")
                    self.__table = table
                    self.__index_by_name = table.get_column_indexes_by_string_content()
                    self.__table_rows_iter = iter(self.__table.rows)
                
                def __next__(self):
                    row = next(self.__table_rows_iter)
                    return {name: row.get_cell(index).value for name, index in self.__index_by_name.items()}
            
            return TableWithHeader2DictGenerator(table)
        else:
            # index_by_name = table.get_column_indexes_by_string_content()
            #
            # res = []
            # for row in table.rows:
            #     new_dict = {name: row.get_cell(index).value for name, index in index_by_name.items()}
            #     res.append(new_dict)
            #
            # return res
            gen = cls.convert_table_with_header_to_dict_list(table, as_generator=True)
            return [e for e in gen]
    
    @classmethod
    def convert_table_2_list_of_tuples(cls, table, as_generator=False):
        if not ValueTableManager.is_value_table(table):
            return TableConverter.convert_table_2_list_of_tuples(table, as_generator=as_generator)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting table to a list of tuples...")
        if as_generator:
            class Table2TupleGenerator(BaseGenerator):
                def __init__(self, table):
                    super().__init__(name="table with header to dict generator")
                    self.__table = table
                    self.__table_rows_iter = iter(self.__table.rows)
                
                def __next__(self):
                    row = next(self.__table_rows_iter)
                    return tuple((cell.value for cell in row.cells))
            
            return Table2TupleGenerator(table)
        else:
            # return [tuple((cell.content for cell in row.cells)) for row in table.rows]
            gen = cls.convert_table_2_list_of_tuples(table, as_generator=True)
            return [e for e in gen]
    
    @classmethod
    def convert_name_value_table_2_dict(cls, table, with_original_value_content = False):
        if not ValueTableManager.is_value_table(table):
            return TableConverter.convert_name_value_table_2_dict(table)
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting Name/Value table to dict (table = {table})")
        
        res = {}
        if table is not None:
            # Verify table structure
            ValueTableManager.verify_table_is_name_value_table(table)
            ValueTableManager.verify_is_value_table(table)
            
            for row in table.rows:
                if row.get_cell(1).value_type not in [ValueTypes.NotApplicable]:
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace("Converting row (%s)", row)
                    
                    name = row.get_cell(0).value
                    if with_original_value_content:
                        value = row.get_cell(1).content
                    else:
                        value = row.get_cell(1).value
                    
                    if name in res:
                        raise FunctionalException("Name '{}' appears several times in table".format(name))
                    else:
                        res[name] = value
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting Name/Value table to dict (table = {table}) => {res}")
        return res
        
    @classmethod
    def convert_name_value_table_2_list_and_dict(cls, table, with_original_value_content = False):
        # if not ValueTableManager.is_value_table(table):
        #     return TableConverter.convert_name_value_table_2_list_and_dict(table)
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting Name/Value table to list and dict (table = {table})")

        # Verify table structure
        ValueTableManager.verify_table_is_name_value_table(table)
        ValueTableManager.verify_is_value_table(table)
        
        res_list = []
        res_dict = {}
        for row in table.rows:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace("Converting row (%s)", row)
            
            name = row.get_cell(0).value
            if with_original_value_content:
                value = row.get_cell(1).content
            else:
                value = row.get_cell(1).value
            
            if name and len(name) > 0:
                if name in res_dict:
                    raise FunctionalException("Name '{}' appears several times in table".format(name))
                else:
                    res_dict[name] = value
            else:
                res_list.append(value)
                
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting Name/Value table to list and dict (table = {table}) => {res_list} ; {res_dict}")
        return [res_list, res_dict]
 
    @classmethod
    def convert_name_value_table_2_json_object(cls, table, converter=None):
        if not ValueTableManager.is_value_table(table):
            return TableConverter.convert_name_value_table_2_json_object(table, converter=converter)
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting Name/Value table to json object (table = {table})")
        if converter is None:
            converter = JsonConverter()
        
        # Verify table structure
        ValueTableManager.verify_table_is_name_value_table(table)
        ValueTableManager.verify_is_value_table(table)
        
        res = {}
        for row in table.rows:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace("Converting row (%s)", row)
            
            if row.get_cell(1).value_type in [ValueTypes.NotApplicable, ValueTypes.Merged]:
                pass
            elif ValueTypes.is_string(row.get_cell(1).value_type):
                name = row.get_cell(0).value
                value = row.get_cell(1).value
                
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                
                    logger.trace(f"Adding for name '{name}' string value [{value}] (type: {Typing.get_object_class_fullname(value)})")
                cls._fill_json_dict(res, name, value)
            else:
                name = row.get_cell(0).value
                value = row.get_cell(1).value
                value = converter.to_json(value)
                
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                
                    logger.trace(f"Adding for name '{name}' json value [{value}] (type: {Typing.get_object_class_fullname(value)})")
                cls._fill_json_dict(res, name, value)
                
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting Name/Value table to json object (table = {table}) => {res}")
        return res

    @classmethod
    def convert_x_name_value_table_2_list_x_values(cls, table, x_name, with_original_value_content = False):
        # if not ValueTableManager.is_value_table(table):
        #     return TableConverter.convert_x_name_value_table_2_list_x_values(table, x_name)
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting {x_name}/Name/Value table to list of {x_name} and values dict (table = {table})")

        # Verify table structure
        ValueTableManager.verify_table_is_x_name_value_table(table, x_name)
        
        x_list = []
        res_dict = {}
        for row in table.rows:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace("Converting row (%s)", row)

            x = row.get_cell(0).value
            name = row.get_cell(1).value
            if with_original_value_content:
                value = row.get_cell(2).content
            else:
                value = row.get_cell(2).value
            
            if x not in res_dict:
                res_dict[x] = {}
                x_list.append(x)
            res_dict[x][name] = value
                
        res = [{x_name:x, 'values':res_dict[x]} for x in x_list]
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting {x_name}/Name/Value table to list of {x_name} and values dict (table = {table}) => {res}")
        return res
        
    
    
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
import selectors
from holado_python.standard_library.socket.non_blocking_socket import TCPNonBlockingSocketClient
import copy
import threading
from holado_python.standard_library.socket.blocking_socket import TCPBlockingSocketClient
from holado_core.common.tools.tools import Tools
from holado.common.handlers.undefined import undefined_argument

logger = logging.getLogger(__name__)


##########################################################################
## Clients
##########################################################################


class MessageSocketClient(object):
    """
    Base class for message socket client.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, separator=b'\n'):
        super().__init__()
        
        self.__separator = separator
        self.__messages_lock = threading.Lock()
        self.__messages = []

    @property
    def messages(self):
        with self.__messages_lock:
            return copy.copy(self.__messages)
    
    @property
    def nb_messages(self):
        with self.__messages_lock:
            return len(self.__messages)
    
    def _extract_messages_from_data(self, data):
        while True:
            ind = data.in_bytes.find(self.__separator)
            if ind >=0:
                msg = data.in_bytes[:ind]
                self._add_message(msg)
                data.in_bytes = data.in_bytes[ind+len(self.__separator):]
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[{self.name}] New message (total: {self.nb_messages}): {msg}")
                elif Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[{self.name}] New message (total: {self.nb_messages})")
            else:
                break
        
    def _add_message(self, msg):
        with self.__messages_lock:
            self.__messages.append(msg)
    
    def reset_received_messages(self):
        with self.__messages_lock:
            self.__messages.clear()
        
    def read_message(self):
        res = None
        with self.__messages_lock:
            if len(self.__messages) > 0:
                res = self.__messages.pop(0)
        return res
    
    def read_messages(self):
        with self.__messages_lock:
            res = copy.copy(self.__messages)
            self.__messages.clear()
        return res
    
    def write_message(self, msg_bytes):
        self.write(msg_bytes + self.__separator)



class MessageTCPBlockingSocketClient(TCPBlockingSocketClient, MessageSocketClient):
    """
    Message socket client with TCP blocking socket.
    """
    def __init__(self, separator=b'\n', *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument, do_run_with_recv=True, do_run_with_send=True):
        MessageSocketClient.__init__(self, separator)
        TCPBlockingSocketClient.__init__(self, name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay, do_run_with_recv=do_run_with_recv, do_run_with_send=do_run_with_send)

    def _process_recv_send(self, *, read_bufsize=1024, read_kwargs=None, write_kwargs=None):
        # Read from socket & write in socket
        super()._process_recv_send(read_bufsize=read_bufsize, read_kwargs=read_kwargs, write_kwargs=write_kwargs)
        
        # Extract messages from received data
        if self.is_run_with_recv:
            with self._data_lock:
                self._extract_messages_from_data(self._data)



class MessageTCPNonBlockingSocketClient(TCPNonBlockingSocketClient, MessageSocketClient):
    """
    Message socket client with TCP non-blocking socket.
    """
    def __init__(self, separator=b'\n', *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument, do_run_with_recv=True, do_run_with_send=True):
        MessageSocketClient.__init__(self, separator)
        TCPNonBlockingSocketClient.__init__(self, name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay, do_run_with_recv=do_run_with_recv, do_run_with_send=do_run_with_send)

    def _service_connection(self, key, mask, *, read_bufsize=1024, read_kwargs=None, write_kwargs=None):
        # Read from socket & write in socket
        res = super()._service_connection(key, mask, read_bufsize=read_bufsize, read_kwargs=read_kwargs, write_kwargs=write_kwargs)
        
        # Extract messages from received data
        if mask & selectors.EVENT_READ:
            with self._data_lock:
                self._extract_messages_from_data(self._data)
        
        return res










import os
import sys
import shutil

here = os.path.abspath(os.path.dirname(__file__))

test_path = os.path.join(here, "..", "..")
holado_path = os.path.join(test_path, "..", "..", "..")
holado_src_path = os.path.normpath(os.path.join(holado_path, "src"))
sys.path.insert(0, holado_src_path)
print(f"Inserted path: {holado_src_path}")

from holado_core.common.tools.path_manager import PathManager
from holado_protobuf.ipc.protobuf.protobuf_compiler import ProtobufCompiler
from holado_grpc.ipc.rpc.grpc_compiler import GRpcCompiler

def compile_protobuf_proto(remove_destination=True):
    # Define protoc path
#    protoc_exe_path = os.path.join(holado_path, "dependencies", "protoc", "protoc-3.17.3-linux-x86_64", "bin", "protoc")
    # protoc_exe_path = os.path.join(holado_path, "dependencies", "protoc", "protoc-21.5-linux-x86_64", "bin", "protoc")
    # protoc_exe_path = os.path.join(holado_path, "dependencies", "protoc", "protoc-21.12-linux-x86_64", "bin", "protoc")
    protoc_exe_path = os.path.join(holado_path, "dependencies", "protoc", "protoc-23.4-linux-x86_64", "bin", "protoc")
    
    # Define protobuf folder
    resources_proto_path = os.path.normpath(os.path.join(test_path, "resources", "proto"))
    
    # Define proto and generated paths
    proto_path = os.path.join(resources_proto_path, "definitions")
    destination_path = os.path.join(resources_proto_path, "generated")
    
    # Remove existing destination
    if remove_destination and os.path.exists(destination_path) and os.path.isdir(destination_path):
        shutil.rmtree(destination_path)
    
    protoc = ProtobufCompiler()
    protoc.protoc_exe_path = protoc_exe_path
    
    protoc.register_proto_path(os.path.join(proto_path, "protobuf"), os.path.join(destination_path, "protobuf"), os.path.join(proto_path, "protobuf", "protobuf.dev"))
    protoc.register_proto_path(os.path.join(proto_path, "protobuf"), os.path.join(destination_path, "protobuf"), os.path.join(proto_path, "protobuf", "custom_types"))
    
    protoc.compile_all_proto()

def compile_grpc_proto(remove_destination=True):
    # Define protobuf folder
    api_grpc_path = os.path.normpath(os.path.join(test_path, "tools", "django", "api_grpc", "api_grpc", "api1"))
    
    # Define proto and generated paths
    proto_path = os.path.join(api_grpc_path, "proto")
    destination_path = os.path.join(api_grpc_path, "proto")
    
    # Remove existing destination
    if remove_destination and os.path.exists(destination_path) and os.path.isdir(destination_path):
        if destination_path != proto_path:
            shutil.rmtree(destination_path)
        else:
            path_manager = PathManager()
            glob_pattern = os.path.join(destination_path, "*.py")
            path_manager.remove_paths(glob_pattern)
    
    protoc = GRpcCompiler()
    
    protoc.register_proto_path(proto_path, destination_path)
    
    protoc.compile_all_proto()

if __name__ == "__main__":
    compile_protobuf_proto()
    compile_grpc_proto(remove_destination=False)
    
    
